# **************************************************************************
# *
# * Authors:     David Herreros Calero (dherreros@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************


import os
import sys
from functools import partial
from xmipp_metadata.image_handler import ImageHandler

import numpy as np
from multiprocessing import Process

from watchdog.observers import Observer
from watchdog.events import LoggingEventHandler

from sklearn.neighbors import KDTree
from sklearn.cluster import KMeans

from bokeh.models import HoverTool, LinearColorMapper, PointDrawTool, ColumnDataSource
from bokeh.plotting import figure
from bokeh.models import Select, Button, TextInput
from bokeh.layouts import column, layout
from bokeh.server.server import Server
from bokeh.util.hex import hexbin
from bokeh.transform import linear_cmap
from bokeh.palettes import all_palettes

from PyQt5 import QtCore, QtWidgets, QtWebEngineWidgets, QtGui
from PyQt5.QtCore import QUrl, QTimer
from PyQt5.QtWidgets import QTabWidget

from traits.api import HasTraits, Instance, Array, Float, String, on_trait_change
from traitsui.api import View, Item, HGroup, Group, HSplit, RangeEditor

from mayavi import mlab
from mayavi.core.api import PipelineBase
from mayavi.core.ui.api import SceneEditor, MayaviScene, MlabSceneModel

from pyworkflow.utils import getExt

import flexutils


class InteractiveAnnotate2D(QtWidgets.QWidget):

    # ---------------------------------------------------------------------------
    # Init
    # ---------------------------------------------------------------------------
    def __init__(self, **kwargs):
        super(InteractiveAnnotate2D, self).__init__()
        # Attributes from kwargs
        self.class_inputs = kwargs
        self.data = kwargs.get("data")
        self.z_space = kwargs.get("z_space")
        self.path = kwargs.get("path")
        self.mode = kwargs.get("mode")
        self.data_tree_3d = KDTree(self.data)
        self.data_tree_2d = KDTree(self.data[:, :-1])
        self.z_space_tree = KDTree(self.z_space)
        self.adding_kmeans = False
        self.renaming_selection = False
        self.updating_2d = False
        # self.hexbin_bins = ColumnDataSource(data=dict(r=[], q=[], counts=[]))

        # Setup main window
        self.fig = figure(title="HexBin flexible landscape",
                          match_aspect=True, tools="wheel_zoom,reset,pan",
                          background_fill_color='black',
                          toolbar_location="below",
                          plot_width=650, plot_height=600)
        self.fig.grid.visible = False

        # Setup tabs Widget
        self.mayavi_tabs = QTabWidget()

        # Setup Mayavi window
        self.createVWInterface(view="volume_viewer", **kwargs)
        self.createVWInterface(view="landscape", **kwargs)

        # Setup server
        self.server = Server({'/': self.createInterface}, num_procs=1)
        self.server.start()
        # server.io_loop.start()

        # Web engine to render plot
        self.window = QtWidgets.QWidget()
        self.web_engine = QtWebEngineWidgets.QWebEngineView()
        self.web_engine.setFixedWidth(1000)
        self.web_engine.load(QUrl("http://localhost:%s/" % self.server.port))

        # PyQt layout
        window_lay = QtWidgets.QHBoxLayout()
        lay_1 = QtWidgets.QVBoxLayout()
        lay_1.addWidget(self.web_engine)
        lay_2 = QtWidgets.QVBoxLayout()
        lay_2.addWidget(self.mayavi_tabs)
        self.mayavi_tabs.addTab(self.mayavi_landscape_view, "Landscape View")
        self.mayavi_tabs.addTab(self.mayavi_vw_view, "Volume Viewer")
        window_lay.addLayout(lay_1)
        window_lay.addLayout(lay_2)
        # self.window.setFixedSize(1600, 650)
        self.window.setLayout(window_lay)
        self.setWindowIcon(QtGui.QIcon(os.path.join(os.path.dirname(flexutils.__file__), "icon_square.png")))

        # Start Mayavi window
        # self.mayavi_view.visualization.configure_traits()

        # Watcher for Mayavi
        class MayaviHandler(LoggingEventHandler):
            def on_moved(cls, event):
                if "z_space_mayavi.txt" in event.dest_path:
                    self.update_mayavi_map()

        class ChimeraxHandler(LoggingEventHandler):
            def on_moved(cls, event):
                if "z_space_chimerax.txt" in event.dest_path:
                    self.morph_selections_chimerax()

        mayavi_watcher = Observer()
        open(os.path.join(self.path, "z_space_mayavi.txt"), 'w').close()
        mayavi_watcher.schedule(MayaviHandler(), self.path, recursive=True)
        mayavi_watcher.start()

        chimerax_watcher = Observer()
        open(os.path.join(self.path, "z_space_chimerax.txt"), 'w').close()
        chimerax_watcher.schedule(ChimeraxHandler(), self.path, recursive=True)
        chimerax_watcher.start()

        # Refresh main window every half a second to avoid Mayavi taking grab
        self.__timer = QTimer()
        self.__timer.timeout.connect(self.web_engine.update)
        self.__timer.start(500)

    # ---------------------------------------------------------------------------
    # Create interface
    # ---------------------------------------------------------------------------
    def createInterface(self, doc):
        self.bokeh_doc = doc

        # Colormap control
        colormaps = {"Viridis": "Viridis", "Cividis": "Cividis", "Plasma": "Plasma",
                     "Inferno": "Inferno", "BuPu": "BuPu", "Oranges": "Oranges",
                     "PuRd": "PuRd"}
        self.colormap_select = Select(value="Viridis", title='Colormap', options=sorted(colormaps.keys()))
        self.colormap_select.on_change('value', self.update_color_map)

        # Save selections button
        button_save_selections = Button(label="Save current selections", button_type="success")
        button_save_selections.on_click(self.save_selections)

        # Rename selections
        self.label_input = TextInput(value="", title="New label:", width=150)
        button_rename_selection = Button(label="Rename selection", button_type="success", width=50, align="end")
        button_rename_selection.on_click(self.rename_selection)

        # Cluster space
        self.n_clusters = TextInput(value="10", title="Number of clusters:", width=150)
        button_cluster = Button(label="Cluster KMeans", button_type="success", width=50, align="end")
        button_cluster.on_click(self.cluster_space_kmeans)

        # ChimeraX morphing
        path_type = {"Salesman": "Salesman", "Random walk": "Random walk"}
        self.path_select = Select(value="Salesman", title='Path type', options=sorted(path_type.keys()),
                                  width=150)
        button_do_morph = Button(label="Morph Chimerax", button_type="success",
                                 width=50, align="end")
        button_do_morph.on_click(self.save_selections_chimerax)

        # Initiate tools
        self.initiatePointDrawTool()

        # Hexbin plot
        bins = hexbin(self.data[:, 0], self.data[:, 1], 0.1)
        self.hexbin_bins = ColumnDataSource(data=dict(r=[], q=[], counts=[]))
        self.hexbin_bins.data = dict(r=bins.r, q=bins.q, counts=bins.counts)

        # Watcher for Mayavi
        class MayaviHandler(LoggingEventHandler):
            def on_moved(cls, event):
                if os.path.join(self.path, "projected_coords.txt") in event.dest_path:
                    self.update_2d_landscape()

        mayavi_landscape_watcher = Observer()
        open(os.path.join(self.path, "projected_coords.txt"), 'w').close()
        mayavi_landscape_watcher.schedule(MayaviHandler(), self.path, recursive=True)
        mayavi_landscape_watcher.start()

        self.hexbin = self.fig.hex_tile(q='q', r='r', source=self.hexbin_bins, size=0.1, level="underlay", line_color=None,
                                        fill_color=linear_cmap('counts', 'Viridis256', 0, max(self.hexbin_bins.data["counts"])))

        # Layout
        controls_colormap = column(self.colormap_select)
        controls_save_selections = column(button_save_selections)

        # Server
        doc.add_root(layout([[self.fig, [controls_colormap,
                                         [self.label_input, button_rename_selection],
                                         [self.n_clusters, button_cluster],
                                         [self.path_select, button_do_morph],
                                         controls_save_selections
                                         ]
                              ]
                             ]
                            ))

    def createVWInterface(self, view, **kwargs):
        if self.mode == "Zernike3D":
            inputs_mayavi = {"path": self.path,
                             "mode": self.mode,
                             "L1": int(kwargs.get("L1")),
                             "L2": int(kwargs.get("L2"))}
        elif self.mode == "CryoDrgn":
            inputs_mayavi = {"path": self.path,
                             "mode": self.mode,
                             "weights": kwargs.get("weights"),
                             "config": kwargs.get("config"),
                             "boxsize": int(kwargs.get("boxsize")),
                             "sr": float(self.class_inputs["sr"])}
        elif self.mode == "HetSIREN":
            inputs_mayavi = {"path": self.path,
                             "mode": self.mode,
                             "weights": kwargs.get("weights"),
                             "step": kwargs.get("step"),
                             "sr": float(self.class_inputs["sr"])}
        elif self.mode == "NMA":
            inputs_mayavi = {"path": self.path,
                             "mode": self.mode,
                             "weights": kwargs.get("weights"),
                             "sr": float(self.class_inputs["sr"]),
                             "boxsize": int(self.class_inputs["boxsize"])}
        if view == "volume_viewer":
            self.mayavi_vw_view = MayaviQWidget(view=view, **inputs_mayavi)
        elif view == "landscape":
            self.mayavi_landscape_view = MayaviQWidget(view=view, data=kwargs.get("data"), path=kwargs.get("path"))

    # ---------------------------------------------------------------------------
    # Tools
    # ---------------------------------------------------------------------------
    def initiatePointDrawTool(self):
        # Read selections or make default
        # self.seeds = ColumnDataSource({'x': [], 'y': [], 'names': []})
        self.loadSelections()

        # Selection callbacks
        self.seeds.on_change("data", self.set_default_values_selection, self.save_z_mayavi, self.update_selection_indeces)
        self.seeds.selected.on_change('indices', self.save_z_mayavi)

        # Tooltips setup
        TOOLTIPS = [
            ("index", "$index"),
            ("(x,y)", "($x, $y)"),
            ("label", "@names"),
        ]

        # Renderers for selections and tooltips
        self.renderer = self.fig.scatter(x='x', y='y', source=self.seeds, color='white', size=10,
                                        line_color='black', line_width=3)
        draw_tool = PointDrawTool(renderers=[self.renderer])
        self.fig.add_tools(draw_tool)
        self.fig.add_tools(HoverTool(tooltips=TOOLTIPS, renderers=[self.renderer]))

    # ---------------------------------------------------------------------------
    # Callbacks
    # ---------------------------------------------------------------------------
    def update_color_map(self, attrname, old, new):
        palette = all_palettes[new]
        palette = palette[list(palette)[-1]]
        color_mapper = LinearColorMapper(palette=palette, low=min(self.hexbin_bins.data["counts"]),
                                         high=max(self.hexbin_bins.data["counts"]))
        # color_mapper = LinearColorMapper(palette=new, low=np.amin(), high=np.amax())
        self.hexbin.glyph.fill_color['transform'].palette = color_mapper.palette

    def save_selections(self):
        selections = self.seeds.data
        selected_points = np.asarray([selections["x"], selections["y"], selections["z"], selections["idx"]]).T
        selected_z_space = self.z_space[np.asarray(selections["idx"]).reshape(-1), :]

        np.savetxt(os.path.join(self.path, "saved_selections.txt"), selected_z_space)
        np.savetxt(os.path.join(self.path, "selections.txt"), selected_points)
        with open(os.path.join(self.path, 'selection_names.txt'), 'w') as f:
            for line in selections["names"]:
                f.write(line + "\n")

    def rename_selection(self):
        self.renaming_selection = True
        selected = self.seeds.selected.indices
        if len(selected) == 1:
            names = self.seeds.data['names']
            names[selected[0]] = self.label_input.value
            self.seeds.data['names'] = names
        self.renaming_selection = False

    def set_default_values_selection(self, attrname, old, new):
        if not self.adding_kmeans and not self.updating_2d:
            if len(new["names"]) > len(old["names"]):
                names = old["names"]
                names.append("class_%d" % (len(new["names"]) - 1))

                z = old["z"]
                z.append(0.0)

                idx = old["idx"]
                new_point = np.asarray([new["x"][-1], new["y"][-1]]).reshape(1, -1)
                _, ind = self.data_tree_2d.query(new_point, k=1)
                idx.append(ind[0].astype(int).item())

                data = {"x": new["x"], "y": new["y"], "z": z, "names": names,
                        "idx": idx}
                self.seeds.data = data


    def save_z_mayavi(self, attrname, old, new):
        if not self.adding_kmeans and not self.renaming_selection and not self.updating_2d:
            idx = None
            if isinstance(new, dict):
                idx = list(set(new["idx"]) - set(old["idx"]))
                if len(idx) == 0:
                    idx = new["idx"][-1]
                else:
                    idx = idx[-1]
            elif isinstance(new, list) and len(new) == 1:
                idx = self.seeds.data["idx"][new[0]]

                # x, y = self.seeds.data["x"][new[0]], self.seeds.data["y"][new[0]]
                # new_point = np.asarray([x, y]).reshape(1, -1)
                # _, ind = self.data_tree_2d.query(new_point, k=1)
                # idx = ind[0].astype(int).item()

            if idx:
                z_s = self.z_space[int(idx), :]
                np.savetxt(os.path.join(self.path, "temp.txt"), z_s, delimiter=",")
                os.rename(os.path.join(self.path, "temp.txt"),
                          os.path.join(self.path, "z_space_mayavi.txt"))

    def update_selection_indeces(self, attrname, old, new):
        if isinstance(new, dict) and not None in new["x"] and not None in old["x"]:
            x = list(set(new["x"]) - set(old["x"]))
            if len(x) > 0:
                idn = new["x"].index(x[0])
                x, y = new["x"][idn], new["y"][idn]
                new_point = np.asarray([x, y]).reshape(1, -1)
                _, ind = self.data_tree_2d.query(new_point, k=1)
                idx = ind[0].astype(int).item()

                seeds_data = dict(self.seeds.data)
                new_idx = seeds_data["idx"]
                new_idx[idn] = idx
                seeds_data["idx"] = new_idx

                async def update(data):
                    self.seeds.data = data

                self.bokeh_doc.add_next_tick_callback(partial(update, seeds_data))

    def save_selections_chimerax(self):
        selected_data = self.seeds.data
        selected_z_space = self.z_space[np.asarray(selected_data["idx"]).reshape(-1), :]
        sel_names = np.asarray(selected_data["names"])

        with open(os.path.join(self.path, 'selection_names_chimerax.txt'), 'w') as f:
            for line in sel_names:
                f.write(line + "\n")
            f.write(self.path_select.value + "\n")
        np.savetxt(os.path.join(self.path, "temp.txt"), selected_z_space)
        os.rename(os.path.join(self.path, "temp.txt"),
                  os.path.join(self.path, "z_space_chimerax.txt"))

    def update_mayavi_map(self):
        self.mayavi_vw_view.visualization.updateMap()
        # self.web_engine.update()
        # self.mayavi_vw_view.update()

    def cluster_space_kmeans(self):
        n_clusters = int(self.n_clusters.value)
        # First delete all automatic selections associated to a previous KMeans
        new_x, new_y, new_z, new_idx, new_names = [], [], [], [], []
        old_x, old_y, old_z, old_idx, old_names = self.seeds.data["x"], self.seeds.data["y"], \
                                                  self.seeds.data["z"], self.seeds.data["idx"], \
                                                  self.seeds.data["names"]
        for x, y, z, idx, name in zip(old_x, old_y, old_z, old_idx, old_names):
            if not 'kmean' in name:
                new_x.append(x)
                new_y.append(y)
                new_z.append(z)
                new_idx.append(idx)
                new_names.append(name)
        # Compute KMeans and save automatic selection
        if int(n_clusters) > 0:
            clusters = KMeans(n_clusters=int(n_clusters)).fit(self.z_space)
            centers = clusters.cluster_centers_
            _, inds = self.z_space_tree.query(centers, k=1)
            for idx, ind in enumerate(inds):
                new_x.append(self.data[ind[0], 0])
                new_y.append(self.data[ind[0], 1])
                new_z.append(self.data[ind[0], 2])
                new_idx.append(ind[0].astype(int).item())
                new_names.append("kmean_%d" % (idx + 1))
        self.adding_kmeans = True
        new_data = {"x": new_x, "y": new_y, "z": new_z, "idx": new_idx, "names": new_names}
        self.seeds.data = new_data
        self.adding_kmeans = False

    def update_2d_landscape(self):
        if os.stat(os.path.join(self.path, "projected_coords.txt")).st_size != 0:
            data = np.loadtxt(os.path.join(self.path, "projected_coords.txt"))
            bins = hexbin(data[:, 0], data[:, 1], 0.5)
            self.data = data
            self.data_tree_3d = KDTree(data)
            self.data_tree_2d = KDTree(data[:, :-1])

            # selected_data = self.seeds.data
            # selected_points = np.asarray([selected_data["x"], selected_data["y"], selected_data["z"]]).T
            # names = selected_data["names"]
            # W = np.ones((selected_points.shape[0], 1))
            # hmgns_world_coords = np.column_stack((selected_points, W))
            # comb_trans_mat = np.loadtxt("comb_trans_mat.txt")
            # view_coords = np.dot(comb_trans_mat, hmgns_world_coords.T).T
            # norm_view_coords = view_coords / (view_coords[:, 3].reshape(-1, 1))
            # view_to_disp_mat = np.loadtxt("view_to_disp_mat.txt")
            # disp_coords = np.dot(view_to_disp_mat, norm_view_coords.T).T

            selected_data = self.seeds.data
            idx = selected_data["idx"]
            idx_np = np.asarray(idx).astype(int)
            new_data = {"x": data[idx_np, 0].tolist(),
                        "y": data[idx_np, 1].tolist(),
                        "z": data[idx_np, 2].tolist(),
                        "idx": idx,
                        "names": selected_data["names"]}

            async def update(data):
                self.updating_2d = True
                self.seeds.data = data
                self.updating_2d = False

            self.bokeh_doc.add_next_tick_callback(partial(update, new_data))


            async def update(r, q, counts):
                self.hexbin_bins.data = dict(r=r, q=q, counts=counts)
                self.fig.renderers.remove(self.hexbin)
                palette = all_palettes[self.colormap_select.value]
                palette = palette[list(palette)[-1]]
                self.hexbin = self.fig.hex_tile(q='q', r='r', source=self.hexbin_bins, size=0.5, level="underlay",
                                                line_color=None,
                                                fill_color=linear_cmap('counts', palette, 0,
                                                                       max(self.hexbin_bins.data["counts"])))

            self.bokeh_doc.add_next_tick_callback(partial(update, r=bins.r, q=bins.q, counts=bins.counts))
            # color_mapper = LinearColorMapper(palette="Viridis256", low=min(self.hexbin_bins.data["counts"]),
            #                                  high=max(self.hexbin_bins.data["counts"]))
            # self.hexbin.glyph.fill_color['transform'].palette = color_mapper.palette

    def morph_selections_chimerax(self):
        # Morph maps in chimerax based on different ordering methods
        from flexutils.viewers.chimera_viewers.viewer_morph_chimerax import FlexMorphChimeraX
        z_space_selected = np.loadtxt(os.path.join(self.path, "z_space_chimerax.txt"))
        with open(os.path.join(self.path, 'selection_names_chimerax.txt'), 'r') as f:
            sel_names = [line.strip("\n") for line in f.readlines()]
        path_select = sel_names[-1]
        sel_names = sel_names[:-1]
        if len(sel_names) > 1:
            if path_select == "Salesman":
                morph_chimerax = FlexMorphChimeraX(z_space_selected, sel_names, self.mode, self.path,
                                                   **self.class_inputs)
                morph_chimerax.showSalesMan()
            elif path_select == "Random walk":
                morph_chimerax = FlexMorphChimeraX(z_space_selected, sel_names, self.mode, self.path,
                                                   **self.class_inputs)
                morph_chimerax.showRandomWalk()

    # ---------------------------------------------------------------------------
    # Read data
    # ---------------------------------------------------------------------------
    def loadSelections(self):
        if os.path.isfile(os.path.join(self.path, "selections.txt")):
            selections = np.loadtxt(os.path.join(self.path, "selections.txt"))
            with open(os.path.join(self.path, 'selection_names.txt'), 'r') as f:
                names = [line.strip("\n") for line in f.readlines()]
            self.seeds = ColumnDataSource({'x': selections[:, 0].tolist(),
                                           'y': selections[:, 1].tolist(),
                                           'z': selections[:, 2].tolist(),
                                           'names': names,
                                           'idx': selections[:, 3].tolist()})
        else:
            x, y, z, names, idx = [], [], [], [], []
            if "n_vol" in self.class_inputs:
                n_vol = int(self.class_inputs["n_vol"])
                x, y, z = self.data[-n_vol:, 0].tolist(), self.data[-n_vol:, 1].tolist(), self.data[-n_vol:, 2].tolist()
                idx = list(range((self.data.shape[0] - n_vol), self.data.shape[0]))
                names.append("reference")
                for idx in range(1, n_vol):
                    names.append("class_%d" % idx)
            data = {'x': x, 'y': y, 'z': z, 'names': names, 'idx': idx}
            self.seeds = ColumnDataSource(data)


    # ---------------------------------------------------------------------------
    # Launch values
    # ---------------------------------------------------------------------------
    def launchViewer(self):
        # Get application
        app = QtWidgets.QApplication.instance()
        if app is None:
            app = QtWidgets.QApplication(sys.argv)

        # Show interface
        self.window.show()

        # Load Bokeh server
        p = Process(target=self.server.io_loop.start, args=())
        p.start()

        # Execute app
        app.exec_()

        # Stop app and sever
        self.server.io_loop.stop()
        self.server.stop()
        p.terminate()


# -------- Mayavi 3D map ----------
class MapView(HasTraits):

    # Map (actor)
    ipw_map = Instance(PipelineBase)

    # Main view
    scene3d = Instance(MlabSceneModel, ())

    # Visualization style
    threshold = Float(0.0)
    op_min = Float(0.0)
    op_max = Float(1.0)

    # Map
    map = Array()
    generated_map = Array()

    # Output path
    path = String()

    # ---------------------------------------------------------------------------
    def __init__(self, **traits):
        self.class_inputs = traits
        super(MapView, self).__init__(**traits)
        self.map
        self.ipw_map

        # Generate map
        self.mode
        if self.mode == "Zernike3D":
            from flexutils.protocols.xmipp.utils.utils import applyDeformationField
            self.generate_map = applyDeformationField
        elif self.mode == "CryoDrgn":
            import cryodrgn
            from cryodrgn.utils import generateVolumes
            cryodrgn.Plugin._defineVariables()
            self.generate_map = generateVolumes
        elif self.mode == "HetSIREN":
            from flexutils.utils import generateVolumesHetSIREN
            self.generate_map = generateVolumesHetSIREN
        elif self.mode == "NMA":
            from flexutils.utils import generateVolumesDeepNMA
            self.generate_map = generateVolumesDeepNMA

    # ---------------------------------------------------------------------------
    # Default values
    # ---------------------------------------------------------------------------
    def _map_default(self):
        return np.zeros([64, 64, 64])

    # ---------------------------------------------------------------------------
    # Scene activation callbaks
    # ---------------------------------------------------------------------------
    @on_trait_change('scene3d.activated')
    def display_scene3d(self):
        self.scene3d.mlab.view(40, 50)
        self.scene3d.scene.background = (0, 0, 0)

    @on_trait_change('scene3d.activated')
    def display_map(self):
        volume = mlab.contour3d(self.map, color=(1, 1, 1), contours=1, figure=self.scene3d.mayavi_scene)
        volume.contour.auto_contours = False
        volume.contour.auto_update_range = True
        setattr(self, 'ipw_map', volume)

    @on_trait_change("threshold")
    def change_map_contour(self):
        volume = getattr(self, 'ipw_map')
        val_max = np.amax(volume.mlab_source.m_data.scalar_data)
        val_min = np.amin(volume.mlab_source.m_data.scalar_data)
        if val_max != val_min:
            contour = (val_max - val_min) * self.threshold + val_min
            volume.contour.contours = [contour]

    # ---------------------------------------------------------------------------
    # Conversion functions
    # ---------------------------------------------------------------------------
    def getDataAtThreshold(self, data, level):
        return np.asarray(np.where(data >= level))

    # ---------------------------------------------------------------------------
    # Read functions
    # ---------------------------------------------------------------------------
    def readMap(self, file):
        map = ImageHandler().read(file).getData()
        return map

    # ---------------------------------------------------------------------------
    # View functions
    # ---------------------------------------------------------------------------
    def updateMap(self):
        z = np.loadtxt(os.path.join(self.path, "z_space_mayavi.txt"), delimiter=",")
        if self.generate_map is not None:
            if self.mode == "Zernike3D":
                self.generate_map("reference.mrc", "mask.mrc", "deformed.mrc",
                                  self.path, z,
                                  int(self.class_inputs["L1"]), int(self.class_inputs["L2"]), 32)
                self.generated_map = self.readMap(os.path.join(self.path, "deformed.mrc"))
            elif self.mode == "CryoDrgn":
                self.generate_map(z, self.class_inputs["weights"],
                                  self.class_inputs["config"], self.path, downsample=int(self.class_inputs["boxsize"]),
                                  apix=float(self.class_inputs["sr"]))
                self.generated_map = self.readMap(os.path.join(self.path, "vol_000.mrc"))
            elif self.mode == "HetSIREN":
                self.generate_map(self.class_inputs["weights"], z,
                                  self.path, step=self.class_inputs["step"])
                self.generated_map = self.readMap(os.path.join(self.path, "decoded_map_class_01.mrc"))
            elif self.mode == "NMA":
                self.generate_map(self.class_inputs["weights"], z,
                                  self.path, sr=float(self.class_inputs["sr"]), xsize=int(self.class_inputs["boxsize"]))
                self.generated_map = self.readMap(os.path.join(self.path, "decoded_map_class_01.mrc"))

            volume = getattr(self, 'ipw_map')
            val_max = np.amax(volume.mlab_source.m_data.scalar_data)
            val_min = np.amin(volume.mlab_source.m_data.scalar_data)
            volume.mlab_source.reset(scalars=self.generated_map)
            if val_max == val_min:
                volume.contour.contours = [0.0001]

    # ---------------------------------------------------------------------------
    # The layout of the dialog created
    # ---------------------------------------------------------------------------
    view = View(HSplit(
        HGroup(
            Group(
                Item('scene3d',
                     editor=SceneEditor(scene_class=MayaviScene),
                     height=250, width=300),
                Item('threshold',
                     editor=RangeEditor(format='%.02f', low_name='op_min', high_name='op_max', mode='slider'),
                     ),
                show_labels=False
            )
        )
    ),
    resizable=True,
    title='Point cloud clustering',
    icon=os.path.join(os.path.dirname(flexutils.__file__), "icon_square.png")
    )


# -------- Mayavi 3D landscape ----------
class LandscapeView(HasTraits):

    # Map (actor)
    ipw_pc = Instance(PipelineBase)

    # Main view
    scene3d = Instance(MlabSceneModel, ())

    # Visualization style
    opacity = Float(0.0)
    scale_points = Float(0.5)
    op_min = Float(0.0)
    op_max = Float(1.0)
    sc_min = Float(0.0)
    sc_max = Float(10.0)

    # Output path
    path = String()

    # ---------------------------------------------------------------------------
    def __init__(self, **traits):
        self.class_inputs = traits
        super(LandscapeView, self).__init__(**traits)

    # ---------------------------------------------------------------------------
    # Default values
    # ---------------------------------------------------------------------------

    # ---------------------------------------------------------------------------
    # Scene activation callbaks
    # ---------------------------------------------------------------------------
    @on_trait_change('scene3d.activated')
    def display_scene3d(self):
        self.scene3d.mlab.view(40, 50)
        self.scene3d.scene.background = (0, 0, 0)
        self.scene3d.scene.interactor.add_observer("EndInteractionEvent", self.get_projected_coords)

    @on_trait_change('scene3d.activated')
    def display_landscape(self):
        data = self.class_inputs["data"]
        scatter = mlab.pipeline.scalar_scatter(data[:, 0], data[:, 1], data[:, 2],
                                               figure=self.scene3d.mayavi_scene)
        scatter = mlab.pipeline.glyph(scatter,
                                      opacity=self.opacity,
                                      scale_mode='none', scale_factor=.5, mode='sphere', colormap="hot",
                                      figure=self.scene3d.mayavi_scene)
        scatter.actor.actor.pickable = 0
        setattr(self, 'ipw_pc', scatter)
        # self.ipw_pc.add_observer('InteractionEvent', self.get_projected_coords)

    @on_trait_change("opacity")
    def change_opacity_point_cloud(self):
        self.ipw_pc.actor.property.opacity = self.opacity

    @on_trait_change("scale_points")
    def change_scale_point_cloud(self):
        scale_factor = float(self.scale_points)
        pts = getattr(self, 'ipw_pc')
        pts.glyph.glyph.scale_factor = scale_factor

    def get_projected_coords(self, obj, evt):
        data = self.class_inputs["data"]
        path = self.class_inputs["path"]
        W = np.ones((data.shape[0], 1))
        hmgns_world_coords = np.column_stack((data, W))
        comb_trans_mat = self.get_world_to_view_matrix()
        view_coords = self.apply_transform_to_points(hmgns_world_coords, comb_trans_mat)
        # norm_view_coords = view_coords / (view_coords[:, 3].reshape(-1, 1))
        # view_to_disp_mat = self.get_view_to_display_matrix()
        # disp_coords = self.apply_transform_to_points(norm_view_coords, view_to_disp_mat)
        coords = view_coords[:, :-1]
        np.savetxt(os.path.join(path, "temp.txt"), coords)
        np.savetxt(os.path.join(path, "comb_trans_mat.txt"), comb_trans_mat)
        # np.savetxt(os.path.join(path, "view_to_disp_mat.txt"), view_to_disp_mat)
        os.rename(os.path.join(path, "temp.txt"),
                  os.path.join(path, "projected_coords.txt"))

    # ---------------------------------------------------------------------------
    # Conversion functions
    # ---------------------------------------------------------------------------

    # ---------------------------------------------------------------------------
    # Read functions
    # ---------------------------------------------------------------------------

    # ---------------------------------------------------------------------------
    # View functions
    # ---------------------------------------------------------------------------
    def get_world_to_view_matrix(self):
        """returns the 4x4 matrix that is a concatenation of the modelview transform and
        perspective transform. Takes as input an mlab scene object."""

        # The VTK method needs the aspect ratio and near and far clipping planes
        # in order to return the proper transform. So we query the current scene
        # object to get the parameters we need.
        scene_size = tuple(self.scene3d.get_size())
        clip_range = self.scene3d.camera.clipping_range
        aspect_ratio = float(scene_size[0]) / float(scene_size[1])

        # this actually just gets a vtk matrix object, we can't really do anything with it yet
        vtk_comb_trans_mat = self.scene3d.camera.get_composite_projection_transform_matrix(
            aspect_ratio, clip_range[0], clip_range[1])

        # get the vtk mat as a numpy array
        np_comb_trans_mat = vtk_comb_trans_mat.to_array()

        return np_comb_trans_mat

    def get_view_to_display_matrix(self):
        """ this function returns a 4x4 matrix that will convert normalized
            view coordinates to display coordinates. It's assumed that the view should
            take up the entire window and that the origin of the window is in the
            upper left corner"""

        # this gets the client size of the window
        x, y = tuple(self.scene3d.get_size())

        # normalized view coordinates have the origin in the middle of the space
        # so we need to scale by width and height of the display window and shift
        # by half width and half height. The matrix accomplishes that.
        view_to_disp_mat = np.array([[x / 2.0, 0., 0., x / 2.0],
                                     [0., -y / 2.0, 0., y / 2.0],
                                     [0., 0., 1., 0.],
                                     [0., 0., 0., 1.]])

        return view_to_disp_mat

    def apply_transform_to_points(self, points, trans_mat):
        """a function that applies a 4x4 transformation matrix to an of
            homogeneous points. The array of points should have shape Nx4"""

        if not trans_mat.shape == (4, 4):
            raise ValueError('transform matrix must be 4x4')

        if not points.shape[1] == 4:
            raise ValueError('point array must have shape Nx4')

        return np.dot(trans_mat, points.T).T

    # ---------------------------------------------------------------------------
    # The layout of the dialog created
    # ---------------------------------------------------------------------------
    view = View(HSplit(
        HGroup(
            Group(
                Item('scene3d',
                     editor=SceneEditor(scene_class=MayaviScene),
                     height=250, width=300),
                Item('opacity',
                     editor=RangeEditor(format='%.02f', low_name='op_min', high_name='op_max', mode='slider'),
                     ),
                Item('scale_points',
                     editor=RangeEditor(format='%.02f', low_name='sc_min', high_name='sc_max', mode='slider'),
                     ),
                show_labels=False
            )
        )
    ),
    resizable=True,
    title='Point cloud clustering',
    icon=os.path.join(os.path.dirname(flexutils.__file__), "icon_square.png")
    )


# -------- Mayavi QWidget -------
class MayaviQWidget(QtWidgets.QWidget):
    def __init__(self, parent=None, view=None, **kwargs):
        QtWidgets.QWidget.__init__(self, parent)
        layout = QtWidgets.QVBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(0)

        if view == "volume_viewer":
            self.visualization = MapView(**kwargs)
        elif view == "landscape":
            self.visualization = LandscapeView(**kwargs)

        # The edit_traits call will generate the widget to embed.
        self.ui = self.visualization.edit_traits(parent=self,
                                                 kind='subpanel').control
        layout.addWidget(self.ui)
        self.ui.setParent(self)


# -------- Viewer call -------
if __name__ == '__main__':
    import argparse

    # Input parameters
    parser = argparse.ArgumentParser()
    parser.add_argument('--data', type=str, required=True)
    parser.add_argument('--z_space', type=str, required=True)
    parser.add_argument('--path', type=str, required=True)
    parser.add_argument('--mode', type=str, required=True)

    def float_or_str(s):
        try:
            return float(s)
        except ValueError:
            return s

    parsed, unknown = parser.parse_known_args()
    for arg in unknown:
        if arg.startswith(("-", "--")):
            # you can pass any arguments to add_argument
            parser.add_argument(arg.split('=')[0], type=float_or_str)

    args = parser.parse_args()

    # Read and generate data
    data = np.loadtxt(args.data)
    z_space = np.loadtxt(args.z_space)

    # Input
    input_dict = vars(args)
    input_dict['data'] = data
    input_dict['z_space'] = z_space

    # Initialize volume slicer
    viewer = InteractiveAnnotate2D(**input_dict)
    viewer.launchViewer()


# **************************************************************************
# *
# * Authors:  David Herreros Calero (dherreros@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************


import os
import numpy as np

import pyworkflow.protocol.params as params
from pyworkflow.viewer import DESKTOP_TKINTER, WEB_DJANGO, ProtocolViewer
from pyworkflow.utils.process import runJob
import pyworkflow.utils as pwutils

from pwem.viewers import ChimeraView
from pwem.objects import SetOfClasses3D

import flexutils
from flexutils.protocols.protocol_annotate_space import ProtFlexAnnotateSpace
from flexutils.protocols.deprecated.protocol_cluster_space import ProtFlexClusterSpace
from flexutils.utils import getOutputSuffix

import xmipp3


class FlexMorphSalesmanViewer(ProtocolViewer):
    """ Morphing along shortest path defined by the salesman's algorithm """
    _label = 'viewer morph path'
    _targets = [ProtFlexClusterSpace, ProtFlexAnnotateSpace]
    _environments = [DESKTOP_TKINTER, WEB_DJANGO]
    OPEN_FILE = "open %s\n"
    VOXEL_SIZE = "volume #%d voxelSize %s\n"
    VOL_HIDE = "vol #%d hide\n"
    VIEW = "view\n"

    def _defineParams(self, form):
        form.addSection(label='Show deformation')
        form.addParam('doShowMorph', params.LabelParam,
                      label="Display the morphing of the maps along the shortest path")

    def _getVisualizeDict(self):
        return {'doShowMorph': self._doShowMorph}

    def _doShowMorph(self, param=None):
        # Get shortest path
        self.computePath()
        path = self.readPath()

        # Useful parameters
        smprt = self.protocol.reference.get().getSamplingRate()

        scriptFile = self.protocol._getPath('morph_orig_ref_chimera.cxc')
        fhCmd = open(scriptFile, 'w')

        for chid, idx in enumerate(path[0]):
            volFile = os.path.abspath(self.protocol._getExtraPath('path_deformed_%d.mrc' % (idx - 1)))
            fhCmd.write(self.OPEN_FILE % volFile)
            fhCmd.write(self.VOXEL_SIZE % ((chid + 1), str(smprt)))
            fhCmd.write(self.VOL_HIDE % (chid + 1))

        num_vol = len(path[0])
        frames = 20 * num_vol
        step = 1 / frames
        fhCmd.write("volume morph #1-%d frames %d playStep %f \n" % (chid + 1, frames, step))
        fhCmd.write(self.VIEW)
        fhCmd.close()

        view = ChimeraView(scriptFile)
        return [view]

    def computePath(self):
        # Get protocol last output
        last_suffix = int(getOutputSuffix(self.protocol, SetOfClasses3D)) - 1
        if last_suffix == 1:
            last_suffix = ""
        else:
            last_suffix = str(last_suffix)
        output = self.protocol.OUTPUT_PREFIX + last_suffix
        output = getattr(self.protocol, output)

        # Get needed inputs
        reference = self.protocol.reference.get().getFileName()
        mask = self.protocol.mask.get().getFileName()
        if pwutils.getExt(reference) == ".mrc":
            reference += ":mrc"
        if pwutils.getExt(mask) == ".mrc":
            mask += ":mrc"

        # Get Zernike3D coefficients and deformed maps
        for idx, rep in enumerate(output.iterRepresentatives()):
            z_clnm = np.fromstring(rep._xmipp_sphCoefficients.get(), sep=",")

            # Write Zernike3D coefficient
            zernike_file = self.protocol._getExtraPath("z_clnm_vw.txt")
            self.writeZernikeFile(z_clnm, zernike_file)

            # Deform volume
            deformedFile = self.protocol._getExtraPath('path_deformed_%d.mrc' % idx)
            params = '-i %s --mask %s --step 1 --blobr 2 -o %s --clnm %s' % \
                     (reference, mask, deformedFile, zernike_file)
            xmipp3.Plugin.runXmippProgram('xmipp_volume_apply_coefficient_zernike3d', params)

        # Get priors
        num_vol = 1
        volumes = self.protocol.volumes.get()
        if volumes:
            num_vol += volumes.getSize()

        # Run salesman's solver
        program = "python " + os.path.join(os.path.dirname(flexutils.__file__), "viewers", "path_finder_tools", "viewer_salesman_solver.py")
        args = "--coords %s --outpath %s --num_vol %d " \
               % (self.protocol._getExtraPath("saved_selections.txt"), self.protocol._getExtraPath("path.txt"),
                  num_vol)
        runJob(None, program, args)

    def writeZernikeFile(self, coeff, outpath):
        particles = self.protocol.particles.get()
        reference = self.protocol.reference.get()
        l1 = particles.L1.get()
        l2 = particles.L2.get()
        dim = reference.getDim()[0]
        with open(outpath, 'w') as fid:
            fid.write(' '.join(map(str, [l1, l2, 0.5 * dim])) + "\n")
            fid.write(' '.join(map(str, coeff.reshape(-1))) + "\n")

    def readPath(self):
        path = []
        with open(self.protocol._getExtraPath("path.txt")) as f:
            lines = f.readlines()
            for line in lines:
                path.append(np.fromstring(line, dtype=float, sep=','))
        return path


# **************************************************************************
# *
# * Authors:     David Herreros Calero (dherreros@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************


from pyworkflow import NEW
from pyworkflow.protocol.params import PointerParam
from pyworkflow.object import Boolean

from pwem.protocols import ProtAnalysis3D, ProtFlexBase


class ProtFlexAssociateSpace(ProtAnalysis3D, ProtFlexBase):
    """ Associate flexible spaces and info to a different set of particles """

    _label = 'associate flex space to particles'
    _devStatus = NEW
    OUTPUT_PREFIX = 'flexParticles'

    # --------------------------- DEFINE param functions ----------------------
    def _defineParams(self, form):
        form.addSection(label='General parameters')
        form.addParam('particlesFlex', PointerParam, label="Particles with flex info",
                      pointerClass='SetOfParticlesFlex', important=True,
                      help="Particles must have a flexibility information associated (Zernike3D, CryoDrgn...")
        form.addParam('particles', PointerParam, label="Particles to associate",
                      pointerClass="SetOfParticles", important=True,
                      help='Particles to associate flexibility information (space and other attributes)')

    # --------------------------- INSERT steps functions ----------------------
    def _insertAllSteps(self):
        self._insertFunctionStep(self._createOutputStep)

    # --------------------------- STEPS functions -----------------------------
    def _createOutputStep(self):
        particlesFlex = self.particlesFlex.get()
        particles = self.particles.get()

        outSet = self._createSetOfParticlesFlex()
        outSet.copyInfo(particlesFlex)
        outSet.setHasCTF(particlesFlex.hasCTF())

        for particleFlex, particle in zip(particlesFlex.iterItems(), particles.iterItems()):
            particleFlex.setLocation(particle.getLocation())
            outSet.append(particleFlex)

        # Save new output
        name = self.OUTPUT_PREFIX
        args = {}
        args[name] = outSet
        self._defineOutputs(**args)
        self._defineSourceRelation(particlesFlex, outSet)
        self._defineSourceRelation(particles, outSet)

    # --------------------------- INFO functions -----------------------------
    def _summary(self):
        summary = []
        if self.getOutputsSize() >= 1:
            summary.append("Particles successfully associated")
        else:
            summary.append("Output annotated classes not ready yet")
        return summary

    def _methods(self):
        return [
            "Associate flexiblity information to a different set of particles",
        ]

    # ----------------------- VALIDATE functions ----------------------------------------
    def validate(self):
        """ Try to find errors on define params. """
        errors = []
        return errors



# **************************************************************************
# *
# * Authors:     David Herreros Calero (dherreros@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************


import numpy as np
from sklearn.decomposition import PCA

from pyworkflow import NEW
from pyworkflow.protocol.params import PointerParam, StringParam

from pwem.protocols import ProtAnalysis3D, ProtFlexBase
from pwem.objects import ParticleFlex


class ProtFlexFilterPCA(ProtAnalysis3D, ProtFlexBase):
    """ Dimensionality reduction of spaces based on different methods """

    _label = 'filter PCA'
    _devStatus = NEW
    OUTPUT_PREFIX = 'outputParticles'

    # --------------------------- DEFINE param functions ----------------------
    def _defineParams(self, form):
        form.addSection(label='General parameters')
        form.addParam('particles', PointerParam, label="Input particles",
                      pointerClass='SetOfParticlesFlex', important=True,
                      help="Particles must have a flexibility information associated (Zernike3D, HetSIREN, CryoDRGN...")
        form.addParam('keepPCA', StringParam, default="1,2,3", label="PCA components to keep",
                      help="Determines the PCA components to keep (values or ranges are valid inputs).\n"
                           "Example: 1,2,3 / 5:6 / 2: / :5 / 90% are valid entries")


    # --------------------------- INSERT steps functions ----------------------
    def _insertAllSteps(self):
        self._insertFunctionStep(self.filterPCASpace)
        self._insertFunctionStep(self.createOutputStep)

    # --------------------------- STEPS functions -----------------------------
    def filterPCASpace(self):
        particles = self.particles.get()
        keepPCA = self.stringToList(self.keepPCA.get())

        # ********* Get Z space *********
        z_space = []
        for particle in particles.iterItems():
            z_space.append(particle.getZFlex())
        z_space = np.asarray(z_space)
        # ********************

        # Define PCA space
        pca = PCA(n_components=z_space.shape[1]).fit(z_space)
        explained_variance_ratio = pca.explained_variance_ratio_
        z_pca = pca.transform(z_space)

        # PCA mask
        mask = np.zeros(z_pca.shape[1], dtype=float)
        if "%" in self.keepPCA.get():
            cum_sum = np.cumsum(explained_variance_ratio)
            idx = np.argmin(np.abs(cum_sum - keepPCA))
            mask[:idx + 1] = 1
        else:
            mask[keepPCA] = 1

        # Get masked PCA space
        z_pca = z_pca * mask[None, ...]
        explained_variance_ratio = explained_variance_ratio[mask.astype(bool)]

        # Get new Flex space
        z_masked = pca.inverse_transform(z_pca)
        np.savetxt(self._getExtraPath("z_pca.txt"), z_pca)
        np.savetxt(self._getExtraPath("z_masked.txt"), z_masked)
        np.savetxt(self._getExtraPath("explained_variance_ratio.txt"), explained_variance_ratio)

    def createOutputStep(self):
        z_masked = np.loadtxt(self._getExtraPath("z_masked.txt"))

        inputSet = self.particles.get()
        progName = inputSet.getFlexInfo().getProgName()
        partSet = self._createSetOfParticlesFlex(progName=progName)

        partSet.copyInfo(inputSet)
        partSet.setHasCTF(inputSet.hasCTF())
        partSet.setAlignmentProj()

        idx = 0
        for particle in inputSet.iterItems():
            outParticle = ParticleFlex(progName=progName)
            outParticle.copyInfo(particle)

            outParticle.setZFlex(z_masked[idx])

            partSet.append(outParticle)

            idx += 1

        self._defineOutputs(outputParticles=partSet)
        self._defineTransformRelation(self.particles, partSet)

    # --------------------------- UTILS functions ----------------------------
    def stringToList(self, input_str, max_num=100):
        # Remove brackets [] if present
        cleaned_str = input_str.replace('[', '').replace(']', '')

        # Split the cleaned string by both commas and spaces to get all numbers or ranges
        segments = [seg for part in cleaned_str.split(',') for seg in part.split()]

        result = []
        for segment in segments:
            if ':' in segment:  # Handle range
                start, end = segment.split(':')
                start = int(start) if start else 1  # If start is empty, default to 1
                end = int(end) if end else max_num  # If end is empty, default to max_num
                result.extend(range(start - 1, end))  # Add range to result
            elif '%' in segment:
                result.append(0.01 * float(segment[:-1]))
            else:
                result.append(int(segment) - 1)  # Add single number to result

        return result

    # --------------------------- INFO functions -----------------------------
    def _summary(self):
        summary = []
        if self.getOutputsSize() >= 1:
            keepPCA = self.stringToList(self.keepPCA.get())
            explained_variance_ratio = np.loadtxt(self._getExtraPath("explained_variance_ratio.txt"))
            for _, outParticles in self.iterOutputAttributes():
                summary.append("Output *%s*:" % outParticles.getNameId().split('.')[1])
            if "%" in self.keepPCA.get():
                cum_sum = np.cumsum(explained_variance_ratio)
                idx = np.argmin(np.abs(cum_sum - keepPCA))
                keepPCA = range(0, idx)
            for pca, ratio in zip(keepPCA, explained_variance_ratio):
                summary.append(f"PC_{pca + 1} --> {ratio:.2f}%")
            summary.append(f"Total variance explained by kept components --> {100. * explained_variance_ratio.sum()}%")
        else:
            summary.append("Output particles not ready yet")
        return summary

    def _methods(self):
        return [
            "PCA filtering of conformational spaces",
        ]

    def _validate(self):
        errors = []

        # Check number of reduced dimensions
        particles = self.particles.get()
        original_dimensions = len(particles.getFirstItem().getZFlex())
        max_dimensions = max(self.stringToList(self.keepPCA.get())) + 1
        min_dimensions = min(self.stringToList(self.keepPCA.get())) + 1
        if max_dimensions > original_dimensions:
            errors.append(f"The number of dimensions in the reduced space must be smalller than or equal to "
                          f"the number of dimensions in the original space (original dimensions "
                          f"= {original_dimensions}. Currently set to {max_dimensions}."
                          "it is set to {dimensions}")
        if min_dimensions < 1:
            errors.append(f"The minimum PCA dimension to be considered must be 1, but {min_dimensions} was given.")

        return errors






import numpy as np
import matplotlib.pyplot as plt
from matplotlib.widgets import Slider, Button
import seaborn as sns
import pandas as pd

from pyworkflow import Config

sns.set_theme()


class InteractiveHist(object):

    def __init__(self, data, protocol):
        self.data = pd.Series(data, name="data")
        self.protocol = protocol
        self.threshold = None
        self.idx_selected = None
        self.total_selected = None

    def update(self, val):
        # Update threshold value
        self.threshold = val

        # Update images selected
        self.idx_selected = np.argwhere(self.data.values <= self.threshold)

        # Update total selected
        self.total_selected = self.idx_selected.size

        # Update the position of the vertical lines
        self.limit_line.set_xdata([val, val])

        # Interactive title
        self.axs.set_title("Total images selected: %d" % self.total_selected)

        # Redraw the figure to ensure it updates
        self.fig.canvas.draw_idle()

    def _plotButton(self):
        axcreateSubset = plt.axes([0.75, 0.02, 0.2, 0.050])
        # Button does not allow to define text color so
        # I write it directly
        color = Config.SCIPION_MAIN_COLOR if hasattr(Config, "SCIPION_MAIN_COLOR") else "maroon"
        axcreateSubset.text(0.5, 0.5, 'Subset Particles',
                            verticalalignment='center',
                            horizontalalignment='center',
                            transform=axcreateSubset.transAxes, color='white')
        bcreateSubset = Button(axcreateSubset, '',  # leave label empty
                               color=color,
                               hovercolor='maroon')
        bcreateSubset.on_clicked(self.createSubset)
        return bcreateSubset

    def createSubset(self, event):
        # Save selected indexes
        np.savetxt(self.protocol._getExtraPath("selected_idx.txt"), self.idx_selected)

        # Save output in protocol
        self.protocol._createOutput()

    def show(self):
        self.fig, self.axs = plt.subplots(1, 1, figsize=(10, 5))
        self.fig.subplots_adjust(bottom=0.25)

        # Create histogram
        sns.kdeplot(self.data, ax=self.axs, fill=True, linewidth=0, color="green")

        # Set axis labels
        self.axs.set_xlabel("RMSE")
        self.axs.set_ylabel("Density")

        # Create the Slider
        slider_ax = self.fig.add_axes([0.20, 0.1, 0.60, 0.03])
        self.slider = Slider(slider_ax, "Threshold", self.data.min(), self.data.max())

        # Create the Vertical lines on the histogram
        self.limit_line = self.axs.axvline(self.slider.val, color='k')

        # Update threshold value
        self.threshold = self.slider.val

        # Update images selected
        self.idx_selected = np.argwhere(self.data.values <= self.threshold)

        # Update total selected
        self.total_selected = self.idx_selected.size

        # Interactive title
        self.axs.set_title("Total images selected: %d" % self.total_selected)

        # Subset button
        self.bcreateSubset = self._plotButton()  # If not assign to something it might be deleted

        # Slider callback
        self.slider.on_changed(self.update)

        plt.show()




# **************************************************************************
# *
# * Authors:     David Herreros Calero (dherreros@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************


import os

import numpy as np
from sklearn.neighbors import NearestNeighbors

from pyworkflow import NEW
from pyworkflow.protocol.params import PointerParam

from pwem.protocols import ProtAnalysis3D, ProtFlexBase
from pwem.objects import VolumeFlex

import flexutils


class ProtFlexOrderVolumes(ProtAnalysis3D, ProtFlexBase):
    """ Order a series of volumes along the shortest path represneted by their landscape representation """

    _label = 'find volumes order'
    _devStatus = NEW
    OUTPUT_PREFIX = 'orderedVolumes'

    # --------------------------- DEFINE param functions ----------------------
    def _defineParams(self, form):
        form.addSection(label='General parameters')
        form.addParam('inputVolumes', PointerParam, label="Input volumes",
                      pointerClass='SetOfVolumesFlex', important=True,
                      help="Volumes to be ordered")

    # --------------------------- INSERT steps functions ----------------------
    def _insertAllSteps(self):
        self._insertFunctionStep(self.orderVolumes)
        self._insertFunctionStep(self.createOutputStep)

    # --------------------------- STEPS functions -----------------------------
    def orderVolumes(self):
        inputVolumes = self.inputVolumes.get()
        dataPath = self._getExtraPath("z_space.txt")
        outPath = self._getExtraPath("path.txt")

        # Save flex space to file
        flex_space = []
        for volume in inputVolumes.iterItems():
            flex_space.append(volume.getZFlex())
        flex_space = np.asarray(flex_space)

        # Get extreme points
        vertexes = self.getBorderPoints(flex_space,
                                        threshold_distance=0.6 * self.automaticThresholdDistance(flex_space))

        best_distance = np.inf
        best_path = None
        for idx in range(vertexes.shape[0]):
            reordered_coords, index_in_original = self.reorderPoints(flex_space, vertexes, idx)
            np.savetxt(dataPath, reordered_coords)

            # Run salesman's solver
            program = "viewer_salesman_solver.py"
            args = "--coords %s --outpath %s --num_vol 0 " \
                   % (dataPath, outPath)
            program = flexutils.Plugin.getProgram(program)
            self.runJob(program, args, numberOfMpi=1)

            # Read path and distance
            path, distance = self.readPath(outPath)

            if distance < best_distance:
                best_distance = distance
                best_path = self.mapIndicesToOriginal(path, index_in_original, reordered_coords.shape[0])

        # Save best solution
        np.savetxt(outPath, best_path)

    def createOutputStep(self):
        inputVolumes = self.inputVolumes.get()
        outPath = self._getExtraPath("path.txt")

        # Get found path
        path = np.loadtxt(outPath, dtype=int)
        path -= 1

        # Generate ordered volumes
        outVols = self._createSetOfVolumesFlex()
        outVols.copyInfo(inputVolumes)
        volumes = [volume for volume in inputVolumes.iterItems(iterate=False)]
        for objId in path:
            volume = VolumeFlex()
            volume.copyInfo(volumes[objId])
            outVols.append(volume)

        # Save new output
        name = self.OUTPUT_PREFIX
        args = {}
        args[name] = outVols
        self._defineOutputs(**args)
        self._defineSourceRelation(inputVolumes, outVols)


    # --------------------------- UTILS functions ----------------------------
    def readPath(self, file):
        # Open the file for reading
        with open(file, 'r') as file:
            # Read the first line from the file
            line = file.readline().strip()  # Remove any leading/trailing whitespace

            # Split the line into a list of strings, then convert each to an integer
            integer_set = [float(number) for number in line.split(',')]

            # Read the second line from the file
            distance = float(file.readline().strip())

        return np.asarray(integer_set).astype(int), distance

    def getBorderPoints(self, points, n_neighbors=3, threshold_distance=0.5):
        # Fit the model
        nbrs = NearestNeighbors(n_neighbors=n_neighbors).fit(points)
        distances, indices = nbrs.kneighbors(points)

        # Identify points with a mean distance to neighbors greater than a threshold
        border_points = points[np.mean(distances, axis=1) > threshold_distance]

        return border_points

    def reorderPoints(self, original_points, border_points, idx):
        # Find the index of the border point in the original points array
        border_point_to_place_first = border_points[idx]
        index_in_original = np.where((original_points == border_point_to_place_first).all(axis=1))[0][0]

        # Rotate the original_points array so that the border point is the first element
        reordered_points = np.concatenate((original_points[index_in_original:], original_points[:index_in_original]),
                                          axis=0)
        return reordered_points, index_in_original

    def automaticThresholdDistance(self, points, n_neighbors=2):
        # Compute the nearest neighbor distances
        nbrs = NearestNeighbors(n_neighbors=n_neighbors).fit(points)
        distances, _ = nbrs.kneighbors(points)

        # Use the distance to the first nearest neighbor (excluding self)
        nearest_neighbor_distance = distances[:, 1]

        mean_distance = np.mean(nearest_neighbor_distance)
        std_distance = np.std(nearest_neighbor_distance)

        # Example: Set threshold as mean distance plus one standard deviation
        threshold_distance = mean_distance + std_distance

        return threshold_distance

    def mapIndicesToOriginal(self, reordered_indices, start_idx, N):
        """
        Map indices from a reordered array back to their original positions.

        :param reordered_indices: Indices in the reordered array.
        :param start_idx: The index in the original array of the point that was moved to the start.
        :param N: Total number of points in the array.
        :return: A list of indices in their original positions.
        """
        original_indices = [(i + start_idx) % N for i in reordered_indices]
        return original_indices


    # ----------------------- VALIDATE functions ----------------------------------------
    def validate(self):
        """ Try to find errors on define params. """
        errors = []
        return errors

    # --------------------------- INFO functions -----------------------------
    def _summary(self):
        summary = []
        outPath = self._getExtraPath("path.txt")
        if os.path.isfile(outPath):
            path = np.loadtxt(outPath, dtype=int)
            summary.append("Detected order: {}".format(path))
        else:
            summary.append("Ordering volumes...")
        return summary

    def _methods(self):
        return [
            "Ordering of volumes based on the shortest path determined by their conformational landscape representation"
        ]



# **************************************************************************
# *
# * Authors:    David Herreros Calero (dherreros@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'coss@cnb.csic.es'
# *
# **************************************************************************


import os
import numpy as np

from pyworkflow import NEW
import pyworkflow.protocol.params as params
import pyworkflow.utils as pwutils
from pyworkflow.object import Float

from pwem.protocols import ProtAnalysis3D, ProtFlexBase

import flexutils
import flexutils.constants as const


class ProtFlexScoreLandscape(ProtAnalysis3D, ProtFlexBase):
    """
    Scoring and (optional) filtering of landscape samples based on neighbour distance
    """

    _label = 'score/filter landscape'
    _devStatus = NEW

    def _defineParams(self, form):
        form.addSection(label='Input')
        form.addParam('inputParticles', params.PointerParam,
                      pointerClass='SetOfParticlesFlex',
                      label="Input landscape", important=True,
                      help='Input particles with the flexibility landscape to be analyzed')
        form.addParam('fast', params.BooleanParam, default=True, label="Fast NN search?",
                      help="If True, apprixmate neighbour search will be used. This will increase "
                           "performance at the expense of reducing the accuracy of the neighbour search")
        form.addParam("neighbours", params.IntParam, default=10, label="Number of neighbours",
                      help="Number of nearest neighbours to be search for each landscape sample. Smaller "
                           "values will better approximate the locality features of a sample, while global "
                           "values will capture more general landscape features.")
        form.addParam('filter', params.EnumParam, choices=['score', 'filter'],
                       default=0, display=params.EnumParam.DISPLAY_HLIST,
                       label='Operation mode',
                       help='Determine whether to retrieve the landscape will be scored or '
                            'filtered out based on a threshold')
        form.addParam('outliersThreshold', params.FloatParam, default=1,
                      label="Outliers distance threshold", condition='filter == 1',
                      help='Z-Score value from 0 to infinite. Only coordinates with a Z-Score smaller than '
                           'or equal to the threshold will be kept in the output')

    # --------------------------- INSERT steps functions ----------------------
    def _insertAllSteps(self):
        self._insertFunctionStep('detectOutliers')
        self._insertFunctionStep('createOutputStep')

    def detectOutliers(self):
        self.scoreOutliers = {}
        self.scoreOutliers = []
        inputParticles = self.inputParticles.get()

        # Prepare landscape
        landscape_path = self._getExtraPath("landscape.txt")
        landscape = np.asarray([particle.getZFlex() for particle in inputParticles.iterItems()])
        np.savetxt(landscape_path, landscape)

        # Compute Z-Scores
        args = "--i %s --o %s --neighbours %d" % \
               (landscape_path, self._getExtraPath(), self.neighbours.get())
        program = "compute_landscape_zscores.py"
        program = flexutils.Plugin.getProgram(program)
        self.runJob(program, args)

    def createOutputStep(self):
        outliersThreshold = self.outliersThreshold.get()
        inputParticles = self.inputParticles.get()
        progName = inputParticles.getFlexInfo().getProgName()
        outParticles = self._createSetOfParticlesFlex(progName=progName)
        outParticles.copyInfo(inputParticles)
        z_scores = np.loadtxt(self._getExtraPath("z_scores.txt"))

        idx = 0
        for particle in inputParticles.iterItems():
            outParticle = particle.clone()
            outParticle.setObjId(None)
            z_score = z_scores[idx]
            if self.filter.get():
                if outliersThreshold >= z_score:
                    outParticle.outlierScore = Float(z_score)
                    outParticles.append(outParticle)
            else:
                outParticles.outlierScore = Float(z_score)
                outParticles.append(outParticle)
            idx += 1

        self._defineOutputs(outputParticles=outParticles)
        self._defineSourceRelation(inputParticles, outParticles)


    # --------------------------- UTILS functions ----------------------

    # --------------------------- INFO functions ----------------------
    def _methods(self):
        methodsMsgs = []
        if self.filter.get() == 0:
            methodsMsgs.append("*Operation mode*: score")
            filter = False
        else:
            methodsMsgs.append("*Operation mode*: filter")
            filter = True
        methodsMsgs.append("*Score Outliers*: True")
        if filter:
            methodsMsgs.append("    * Outlier threshold: %.2f" % self.outliersThreshold.get())
        return methodsMsgs

    def _summary(self):
        summary = []
        if self.getOutputsSize() >= 1:
            summary.append("Output *%s*:" % self.outputParticles.getNameId().split('.')[1])
            summary.append("    * Number of particles kept: *%s*" % self.outputParticles.getSize())
        else:
            summary.append("Output particles not ready yet.")
        return summary



# **************************************************************************
# *
# * Authors:     David Herreros Calero (dherreros@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************


import os
import numpy as np

from pyworkflow import NEW
from pyworkflow.protocol.params import PointerParam, FloatParam

from pwem.protocols import ProtAnalysis3D, ProtFlexBase
from pwem.objects import ClassStructFlex, AtomStructFlex, SetOfClassesStructFlex, SetOfParticlesFlex

import flexutils
import flexutils.constants as const
from flexutils.utils import getOutputSuffix


class XmippProtClusterStructuresZernike3D(ProtAnalysis3D, ProtFlexBase):
    """ Automatic clustering at atomic structure level based on a threshold distance """

    _label = 'structure clustering - Zernike3D'
    _devStatus = NEW
    OUTPUT_PREFIX = 'clusteredStructures'

    # --------------------------- DEFINE param functions ----------------------
    def _defineParams(self, form):
        form.addSection(label='General parameters')
        form.addParam('particles', PointerParam, label="Zernike3D particles",
                      pointerClass='SetOfParticlesFlex', important=True)
        form.addParam('structure', PointerParam, label="Atomic structure",
                      pointerClass="AtomStruct", allowsNull=True,
                      help="Reference atomic structure that should be traced (or at least "
                           "aligned) with the reference map used during the Zernike3D analysis")
        form.addParam('dist', FloatParam, default=5., label="Distance threshold",
                      help="Determine the minimum RMSD between the representative structures "
                           "of each cluster. The number of clusters found by the algorithm will "
                           "depend on the distance threshold: the smaller the threshold, the larger "
                           "the number of clusters that will be retrieved. Units are in Angstroms.")

    # --------------------------- INSERT steps functions ----------------------
    def _insertAllSteps(self):
        self._insertFunctionStep(self.computeClusters)
        self._insertFunctionStep(self.createOutputStep)

    # --------------------------- STEPS functions -----------------------------
    def computeClusters(self):
        # Input data
        particles = self.particles.get()
        structure = self.structure.get().getFileName()
        dist_thr = self.dist.get()
        boxSize = particles.getXDim()
        sr = particles.getSamplingRate()
        L1 = particles.getFlexInfo().L1.get()
        L2 = particles.getFlexInfo().L2.get()

        # Retrieve Zernike3D coefficients
        z_space_part = [particle.getZFlex() for particle in particles.iterItems()]
        z_space_part = sr * np.asarray(z_space_part)
        z_space_file = self._getExtraPath("z_space.txt")
        np.savetxt(z_space_file, z_space_part)

        # Compute clusters
        args = "--z_space % s --pdb %s --boxSize %f --sr %f --distThr %f " \
               "--L1 %d --L2 %d --odir %s" \
               % (z_space_file, structure, boxSize, sr, dist_thr, L1, L2, self._getExtraPath())
        program = "structure_rmsd_clustering.py"
        program = flexutils.Plugin.getProgram(program)
        self.runJob(program, args)

    def createOutputStep(self):
        particles = self.particles.get()
        partIds = list(particles.getIdSet())
        rep_path = self._getExtraPath("clustered_structures/")
        img_rep_path = self._getExtraPath("cluster_img_ids/")
        num_classes = len([name for name in os.listdir(rep_path) if os.path.isfile(os.path.join(rep_path, name))])

        # Create SetOfClasses3D
        suffix = getOutputSuffix(self, SetOfClassesStructFlex)
        classes = self._createSetOfClassesStructFlex(particles, suffix=suffix,
                                                     progName=const.ZERNIKE3D)

        # Popoulate SetOfClassesStructFlex with KMean particles
        for clInx in range(num_classes):

            newClass = ClassStructFlex(progName=const.ZERNIKE3D)
            newClass.copyInfo(particles)
            newClass.setHasCTF(particles.hasCTF())
            representative = AtomStructFlex(progName=const.ZERNIKE3D)
            representative.setFileName(os.path.join(rep_path, "cluster_%d.pdb" % clInx))

            newClass.setRepresentative(representative)

            classes.append(newClass)

            enabledClass = classes[newClass.getObjId()]
            enabledClass.enableAppend()

            img_rep_ids = np.loadtxt(os.path.join(img_rep_path, "cluster_%d.txt" % clInx))

            if img_rep_ids.size > 1:
                for img_idx in img_rep_ids:
                    item = particles[partIds[int(img_idx)]]
                    enabledClass.append(item)
            else:
                item = particles[partIds[int(img_rep_ids)]]
                enabledClass.append(item)

            classes.update(enabledClass)

        # Save new output
        name = self.OUTPUT_PREFIX + suffix
        args = {}
        args[name] = classes
        self._defineOutputs(**args)
        self._defineSourceRelation(particles, classes)

    # --------------------------- INFO functions -----------------------------
    def _summary(self):
        summary = []
        if self.getOutputsSize() >= 1:
            for _, outClasses in self.iterOutputAttributes():
                summary.append("Output *%s*:" % outClasses.getNameId().split('.')[1])
                summary.append("    * Total number of clusters: *%s*" % outClasses.getSize())
        else:
            summary.append("Output clusters not ready yet")
        return summary

    def _methods(self):
        return [
            "Automatic RMSD based clustering of Zernike3D atomic structures",
        ]

    # ----------------------- VALIDATE functions ----------------------------------------
    def validate(self):
        """ Try to find errors on define params. """
        errors = []
        particles = self.particles.get()
        if isinstance(particles, SetOfParticlesFlex):
            if particles.getFlexInfo().getProgName() != const.ZERNIKE3D:
                errors.append("The flexibility information associated with the particles is not "
                              "coming from the Zernike3D algorithm. Please, provide a set of particles "
                              "with the correct flexibility information.")
        return errors



# **************************************************************************
# *
# * Authors:     David Herreros Calero (dherreros@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************


import sys
import os
import subprocess
import time

import psutil
import socket
from contextlib import closing

from PyQt5 import QtWidgets, QtWebEngineWidgets, QtGui
from PyQt5.QtCore import QUrl

from pyworkflow.utils import runJob

import flexutils


class ViewTensorboard(QtWidgets.QWidget):

    # ---------------------------------------------------------------------------
    # Init
    # ---------------------------------------------------------------------------
    def __init__(self, logdir_path):
        super(ViewTensorboard, self).__init__()
        # Setup Tensorboard
        port = self.launchTensorboard(logdir_path)

        # Web engine to render plot
        self.window = QtWidgets.QWidget()
        self.web_engine = QtWebEngineWidgets.QWebEngineView()
        # self.web_engine.setFixedWidth(1000)
        self.web_engine.load(QUrl(f"http://localhost:{port}/"))

        # PyQt layout
        window_lay = QtWidgets.QHBoxLayout()
        lay_1 = QtWidgets.QVBoxLayout()
        lay_1.addWidget(self.web_engine)
        window_lay.addLayout(lay_1)
        self.window.setBaseSize(1600, 650)
        self.window.setLayout(window_lay)
        self.setWindowIcon(QtGui.QIcon(os.path.join(os.path.dirname(flexutils.__file__), "icon_square.png")))

    # ---------------------------------------------------------------------------
    # Create interface
    # ---------------------------------------------------------------------------
    def launchTensorboard(self, logdir_path):
        def is_port_in_use(port: int) -> bool:
            with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
                return s.connect_ex(('localhost', port)) == 0

        with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
            s.bind(('', 0))
            s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            port = s.getsockname()[1]
        program = flexutils.Plugin.getTensorflowProgram("tensorboard", python=False)
        args = f" --logdir {logdir_path} --port {port}"
        self.tensorboard_process = subprocess.Popen(program + args, shell=True, stdout=subprocess.DEVNULL,
                                                    stderr=subprocess.DEVNULL)

        while not is_port_in_use(port):
            time.sleep(0.1)
        return port

    # ---------------------------------------------------------------------------
    # Tools
    # ---------------------------------------------------------------------------

    # ---------------------------------------------------------------------------
    # Callbacks
    # ---------------------------------------------------------------------------

    # ---------------------------------------------------------------------------
    # Read data
    # ---------------------------------------------------------------------------

    # ---------------------------------------------------------------------------
    # Launch values
    # ---------------------------------------------------------------------------
    def launchViewer(self):
        # Show interface
        self.window.show()

def kill(proc_pid):
    process = psutil.Process(proc_pid)
    for proc in process.children(recursive=True):
        proc.kill()
    process.kill()

# -------- Viewer call -------
if __name__ == '__main__':
    import argparse

    # Input parameters
    parser = argparse.ArgumentParser()
    parser.add_argument('--logdir_path', type=str, required=True)

    # Input
    args = parser.parse_args()
    input_dict = vars(args)

    # Get application
    app = QtWidgets.QApplication.instance()
    if app is None:
        app = QtWidgets.QApplication(sys.argv)

    # Initialize volume slicer
    viewer = ViewTensorboard(**input_dict)

    # Execute app
    viewer.launchViewer()
    app.exec_()

    # Close tensorboard process
    kill(viewer.tensorboard_process.pid)




# **************************************************************************
# *
# * Authors:     David Herreros Calero (dherreros@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************


import glob

import numpy as np
import os

from xmipp_metadata.image_handler import ImageHandler

from pyworkflow.object import String, Integer, Float
from pyworkflow import VERSION_2_0
from pyworkflow.utils import removeExt, moveFile, removeBaseExt, getExt
import pyworkflow.protocol.params as params

from pwem.protocols import ProtAnalysis3D, ProtFlexBase
from pwem.objects import SetOfVolumes, Volume, VolumeFlex

import xmipp3

import flexutils.constants as const
from flexutils.utils import readZernikeFile, getXmippFileName
import flexutils


class XmippProtComputeHeterogeneityPriorsZernike3D(ProtAnalysis3D, ProtFlexBase):
    """ Compute Zernike3D priors and assign them to a SetOfVolumes """
    _label = 'compute heterogeneity priors - Zernike3D'
    _lastUpdateVersion = VERSION_2_0
    OUTPUT_SUFFIX = '_%d_crop.mrc'
    ALIGNED_VOL = 'vol%dAligned.mrc'
    OUTPUT_PREFIX = "zernikeVolumes"

    def __init__(self, **args):
        ProtAnalysis3D.__init__(self, **args)
        self.stepsExecutionMode = params.STEPS_PARALLEL

    # --------------------------- DEFINE param functions --------------------------------------------
    def _defineParams(self, form):
        form.addSection(label='Input')
        form.addParam('inputVolumes', params.MultiPointerParam,
                      pointerClass='SetOfVolumes,Volume',
                      label="Input volume(s)", important=True,
                      help='Volumes used to compute the Zernike3D priors')
        form.addParam('reference', params.PointerParam,
                      pointerClass='Volume',
                      label="Zernike3D reference map", important=True,
                      help='Priors computed will be refered to this reference map')
        form.addParam('mask', params.PointerParam,
                      pointerClass='VolumeMask',
                      label="Reference map mask",
                      help='Mask determining where to compute the deformation field in the '
                           'reference volume. The tightest the mask, the higher the '
                           'performance boost')
        form.addParam('boxSize', params.IntParam, default=128,
                      label='Downsample particles to this box size', expertLevel=params.LEVEL_ADVANCED,
                      help='In general, downsampling the volumes will increase performance without compromising '
                           'the estimation the deformation field for each particle. Note that output particles will '
                           'have the original box size, and Zernike3D coefficients will be modified to work with the '
                           'original size volumes')
        form.addParam('l1', params.IntParam, default=3,
                      label='Zernike Degree',
                      expertLevel=params.LEVEL_ADVANCED,
                      help='Degree Zernike Polynomials of the deformation=1,2,3,...')
        form.addParam('l2', params.IntParam, default=2,
                      label='Harmonical Degree',
                      expertLevel=params.LEVEL_ADVANCED,
                      help='Degree Spherical Harmonics of the deformation=1,2,3,...')

    # --------------------------- INSERT steps functions --------------------------------------------
    def _insertAllSteps(self):
        self._insertFunctionStep(self.convertInputStep)
        self._insertFunctionStep(self.computePriorsStep)
        self._insertFunctionStep(self.createOutputStep)

    # --------------------------- STEPS functions ---------------------------------------------------
    def convertInputStep(self):
        reference_file = self.reference.get().getFileName()
        dim = self.reference.get().getDim()[0]
        boxSize = self.boxSize.get()

        # Get input files
        input_files = []
        for pointer in self.inputVolumes:
            obj = pointer.get()
            if isinstance(obj, Volume):
                input_files.append(obj.getFileName())
            elif isinstance(obj, SetOfVolumes):
                for input_map in obj.iterItems():
                    input_files.append(input_map.getFileName())

        # Rigid fitting of maps to reference
        for input_file in input_files:
            output_file = self._getExtraPath(removeBaseExt(input_file) + "_aligned.mrc")
            args = "--i %s --r %s --o %s" % (input_file, reference_file, output_file)
            program = "align_maps.py"
            program = flexutils.Plugin.getProgram(program)
            self.runJob(program, args, numberOfMpi=1)

            # Resize inputs if needed
            if dim != boxSize:
                self.runJob("xmipp_image_resize",
                            "-i %s -o %s --dim %d "
                            % (output_file + ":mrc", output_file, boxSize), numberOfMpi=1,
                            env=xmipp3.Plugin.getEnviron())

        input_mask = getXmippFileName(self.mask.get().getFileName())
        out_mask = self._getExtraPath("ref_mask.mrc")
        out_reference = self._getExtraPath("ref.mrc")
        ih = ImageHandler()
        ih.convert(input_mask, out_mask)
        ih.convert(getXmippFileName(reference_file), out_reference)
        if dim != boxSize:
            self.runJob("xmipp_image_resize",
                        "-i %s --dim %d --interp nearest" % (out_mask, boxSize), numberOfMpi=1,
                        env=xmipp3.Plugin.getEnviron())
            self.runJob("xmipp_image_resize",
                        "-i %s --dim %d" % (out_reference, boxSize), numberOfMpi=1,
                        env=xmipp3.Plugin.getEnviron())

    def computePriorsStep(self):
        reference_file = self._getExtraPath("ref.mrc")
        input_files = glob.glob(self._getExtraPath("*_aligned.mrc"))
        L1 = self.l1.get()
        L2 = self.l2.get()

        for idf, input_file in enumerate(input_files):
            output_file = self._getExtraPath("deformed_%d.mrc" % (idf + 1))
            args = "--i %s --r %s --o %s --l1 %d --l2 %d" \
                   % (reference_file, input_file, output_file, L1, L2)
            program = "find_z_clnm_map.py"
            program = flexutils.Plugin.getProgram(program)
            self.runJob(program, args, numberOfMpi=1)

            z_clnm_file = self._getExtraPath("z_clnm.txt")
            rmsd_def_file = self._getExtraPath("rmsd_def.txt")

            moveFile(z_clnm_file, removeExt(z_clnm_file) + "_%d.txt" % (idf + 1))
            moveFile(rmsd_def_file, removeExt(rmsd_def_file) + "_%d.txt" % (idf + 1))

    def createOutputStep(self):
        reference = self.reference.get()
        mask = self.mask.get()
        dim = reference.getDim()[0]
        sr = reference.getSamplingRate()
        L1 = Integer(self.l1.get())
        L2 = Integer(self.l2.get())
        Rmax = Integer(0.5 * reference.getDim()[0])
        reference_filename = String(reference.getFileName())
        mask_filename = String(mask.getFileName()) if mask else String("")

        zernikeVols = self._createSetOfVolumesFlex(progName=const.ZERNIKE3D)
        zernikeVols.setSamplingRate(sr)
        zernikeVols.getFlexInfo().L1 = L1
        zernikeVols.getFlexInfo().L2 = L2
        zernikeVols.getFlexInfo().Rmax = Rmax
        zernikeVols.getFlexInfo().refMap = reference_filename
        zernikeVols.getFlexInfo().refMask = mask_filename

        input_files = len(glob.glob(self._getExtraPath("*_aligned.mrc")))
        for idf in range(input_files):
            z_clnm_file = self._getExtraPath("z_clnm_%d.txt" % (idf + 1))
            rmsd_def_file = self._getExtraPath("rmsd_def_%d.txt" % (idf + 1))

            basis_params, z_clnm = readZernikeFile(z_clnm_file)
            rmsd_def = np.loadtxt(rmsd_def_file)

            deformation = Float(rmsd_def)
            factor = dim / self.boxSize.get()
            z_clnm = factor * z_clnm
            # Rmax = Float(factor * basis_params[2])

            zernikeVol = VolumeFlex(progName=const.ZERNIKE3D)
            zernikeVol.setFileName(reference.getFileName())
            zernikeVol.setSamplingRate(sr)
            zernikeVol.setZFlex(z_clnm.reshape(-1))
            zernikeVol.getFlexInfo().deformation = deformation
            zernikeVol.getFlexInfo().L1 = L1
            zernikeVol.getFlexInfo().L2 = L2
            zernikeVol.getFlexInfo().Rmax = Rmax
            zernikeVol.getFlexInfo().refMap = reference_filename
            zernikeVol.getFlexInfo().refMask = mask_filename

            zernikeVols.append(zernikeVol)

        # Save new output
        name = self.OUTPUT_PREFIX
        args = {}
        args[name] = zernikeVols
        self._defineOutputs(**args)
        self._defineSourceRelation(reference, zernikeVols)

    # ------------------------- VALIDATE functions -----------------------------
    def validate(self):
        """ Try to find errors on define params. """
        errors = []
        l1 = self.l1.get()
        l2 = self.l2.get()
        if (l1 - l2) < 0:
            errors.append('Zernike degree must be higher than '
                          'SPH degree.')
        return errors







# **************************************************************************
# *
# * Authors:     David Herreros Calero (dherreos@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************


import numpy as np
import os

import pyworkflow.protocol.params as params
from pyworkflow.utils.path import makePath
from pyworkflow import VERSION_2_0

from pwem.protocols import ProtAnalysis3D, ProtFlexBase

import flexutils


class TensorflowProtTrainFlexConsensus(ProtAnalysis3D, ProtFlexBase):
    """ Protocol to train a FlexConsensus network """
    _label = 'train - FlexConsensus'
    _lastUpdateVersion = VERSION_2_0

    # --------------------------- DEFINE param functions --------------------------------------------
    def _defineParams(self, form):
        form.addSection(label='Input')
        form.addHidden(params.USE_GPU, params.BooleanParam, default=True,
                       label="Use GPU for execution",
                       help="This protocol has both CPU and GPU implementation.\
                             Select the one you want to use.")
        form.addHidden(params.GPU_LIST, params.StringParam, default='0',
                       expertLevel=params.LEVEL_ADVANCED,
                       label="Choose GPU IDs",
                       help="Add a list of GPU devices that can be used")
        group = form.addGroup("Data")
        group.addParam('inputSets', params.MultiPointerParam,
                       label="Input particles", pointerClass='SetOfParticlesFlex')
        group = form.addGroup("Latent Space")
        group.addParam('latDim', params.IntParam, default=10, label='Latent space dimension',
                       expertLevel=params.LEVEL_ADVANCED,
                       help="Dimension of the FlexConsensus bottleneck (latent space dimension)")
        form.addSection(label='Network')
        group = form.addGroup("Network hyperparameters")
        group.addParam('stopType', params.EnumParam, choices=['Samples', 'Manual'],
                       default=0, label="How to compute total epochs?",
                       display=params.EnumParam.DISPLAY_HLIST,
                       help="* *Samples*: Epochs will be obtained from the total number of samples "
                            "the network will see\n"
                            "* *Epochs*: Total number of epochs is provided manually")
        group.addParam('epochs', params.IntParam, default=100, condition="stopType==1",
                       label='Number of training epochs')
        group.addParam('maxSamples', params.IntParam, default=1000000, condition="stopType==0",
                       label="Samples",
                       help='Maximum number of samples seen during network training')
        group.addParam('batch_size', params.IntParam, default=64, label='Number of samples in batch',
                      help="Number of samples that will be used simultaneously for every training step. "
                           "We do not recommend to change this value unless you experience memory errors. "
                           "In this case, value should be decreased.")
        group.addParam('lr', params.FloatParam, default=1e-3, label='Learning rate',
                       help="The learning rate determines how fast the network will train based on the "
                            "seen samples. The larger the value, the faster the network although divergence "
                            "might occur. We recommend decreasing the learning rate value if this happens.")
        group = form.addGroup("Extra network parameters")
        group.addParam('split_train', params.FloatParam, default=1.0, label='Traning dataset fraction',
                      help="This value (between 0 and 1) determines the fraction of images that will "
                           "be used to train the network.")
        group = form.addGroup("Logger")
        group.addParam('debugMode', params.BooleanParam, default=False, label='Debugging mode',
                       help="If you experience any error during the training execution, we recommend setting "
                            "this parameter to True followed by a restart of this protocol to generate a more "
                            "informative logging file.")
        form.addParallelSection(threads=4, mpi=0)

    # --------------------------- INSERT steps functions --------------------------------------------
    def _insertAllSteps(self):
        self._insertFunctionStep(self.convertInputStep)
        self._insertFunctionStep(self.trainingStep)

    # --------------------------- STEPS functions ---------------------------------------------------
    def convertInputStep(self):
        data_path = self._getExtraPath("data")
        if not os.path.isdir(data_path):
            makePath(data_path)

        idx = 0
        for pointer_set in self.inputSets:
            particle_set = pointer_set.get()

            # Get data filename
            progName = particle_set.getFlexInfo().getProgName()
            data_file = progName + f"_{idx}.txt"

            # Read flexible space form particles
            z_flex = []
            for particle in particle_set.iterItems():
                z_flex.append(particle.getZFlex())
            z_flex = np.vstack(z_flex)

            # Save flexible space
            np.savetxt(os.path.join(data_path, data_file), z_flex)

            idx += 1

    def trainingStep(self):
        data_path = self._getExtraPath("data")
        out_path = self._getExtraPath()
        network_path = self._getExtraPath("network")
        if not os.path.isdir(network_path):
            os.mkdir(network_path)
        batch_size = self.batch_size.get()
        split_train = self.split_train.get()
        lr = self.lr.get()
        lat_dim = self.latDim.get()
        args = "--data_path %s --out_path %s --lat_dim %d --batch_size %d " \
               "--shuffle --split_train %f --lr %f" \
               % (data_path, out_path, lat_dim, batch_size, split_train, lr)

        if self.stopType.get() == 0:
            args += " --max_samples_seen %d" % self.maxSamples.get()
        else:
            args += " --epochs %d" % self.epochs.get()

        if self.useGpu.get():
            gpu_list = ','.join([str(elem) for elem in self.getGpuList()])
            args += " --gpu %s" % gpu_list

        if self.debugMode.get():
            log_level = 0
        else:
            log_level = 2

        program = flexutils.Plugin.getTensorflowProgram("train_flex_consensus.py", python=False,
                                                        log_level=log_level)
        self.runJob(program, args, numberOfMpi=1)

    # --------------------------- UTILS functions --------------------------------------------

    # --------------------------- INFO functions -----------------------------
    def _summary(self):
        summary = []
        logFile = os.path.abspath(self._getLogsPath()) + "/run.stdout"
        with open(logFile, "r") as fi:
            for ln in fi:
                if ln.startswith("GPU memory has"):
                    summary.append(ln)
                    break
        return summary

    # ----------------------- VALIDATE functions ----------------------------------------



# **************************************************************************
# *
# * Authors:     David Herreros Calero (dherreros@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************


import os
import numpy as np

from pyworkflow import NEW
from pyworkflow.object import CsvList, Boolean
from pyworkflow.protocol import LEVEL_ADVANCED
from pyworkflow.protocol.params import PointerParam, EnumParam, IntParam, BooleanParam, FloatParam, StringParam, \
                                       GPU_LIST, USE_GPU

from pwem.protocols import ProtAnalysis3D, ProtFlexBase
from pwem.objects import ParticleFlex

import flexutils
import flexutils.constants as const


class ProtFlexDimRedSpace(ProtAnalysis3D, ProtFlexBase):
    """ Dimensionality reduction of spaces based on different methods """

    _label = 'dimred space'
    _devStatus = NEW
    OUTPUT_PREFIX = 'outputParticles'

    # --------------------------- DEFINE param functions ----------------------
    def _defineParams(self, form):
        form.addSection(label='General parameters')
        form.addHidden(USE_GPU, BooleanParam, default=True,
                       condition="mode==0 or mode==2",
                       label="Use GPU for execution",
                       help="This protocol has both CPU and GPU implementation.\
                                     Select the one you want to use.")
        form.addHidden(GPU_LIST, StringParam, default='0',
                       expertLevel=LEVEL_ADVANCED,
                       condition="mode==2",
                       label="Choose GPU IDs",
                       help="Add a list of GPU devices that can be used")
        form.addParam('particles', PointerParam, label="Input particles",
                      pointerClass='SetOfParticlesFlex', important=True,
                      help="Particles must have a flexibility information associated (Zernike3D, CryoDrgn...")
        form.addParam('mode', EnumParam, choices=['UMAP', 'PCA', 'deepElastic'],
                      default=0, display=EnumParam.DISPLAY_HLIST,
                      label="Dimensionality reduction method",
                      help="\t * UMAP: usually leads to more meaningfull spaces, although execution "
                           "is higher\n"
                           "\t * PCA: faster but less meaningfull spaces \n"
                           "\t * deepElastic: Variational autencoder based dimred method. This method learns and "
                           "embedding that tries to keep as best as possible the internal clustering structure of "
                           "the original N-D space \n"
                           "UMAP, PCA, and cryoExplode are only computed the first time the are used. Afterwards, they "
                           "will be reused to increase performance.")
        form.addParam('epochs_umap', IntParam, label="Number of UMAP epochs",
                      default=10, condition="mode==0",
                      help="Increasing the number of epochs will lead to more accurate UMAP spaces at the cost "
                           "of larger execution times")
        form.addParam('clusters', IntParam, label="Initial number of clusters", default=10,
                      condition="mode==2",
                      expertLevel=LEVEL_ADVANCED,
                      help="The N-D space will be splitted in the number of cluster specified so the network "
                           "can learn the best way to keep their strucutre in the reduced space.")
        form.addParam('init_power', FloatParam, label="Initial explosion power", default=10.0,
                      condition="mode==2",
                      expertLevel=LEVEL_ADVANCED,
                      help="The initial power to scatter the landscape. This help the network to learn appropiately "
                           "the initial clustering")
        form.addParam('end_power', FloatParam, label="Final explosion power", default=1.0,
                      condition="mode==2",
                      expertLevel=LEVEL_ADVANCED,
                      help="The final power to scatter the landscape. This will determine how close the clusters will "
                           "be in the final embedding.")
        form.addParam('vae_sigma', FloatParam, label="Variational autoencoder sigma", default=1.0,
                      condition="mode==2",
                      expertLevel=LEVEL_ADVANCED,
                      help="Larger values of sigma will enfoce continuity of the final embedding. If set to zero, a "
                           "non-variational autoencoder will be trained.")
        form.addParam('loss_lambda', FloatParam, label="Cosine mapping lambda", default=1.0,
                      condition="mode==2",
                      expertLevel=LEVEL_ADVANCED,
                      help="If 0.0, cosine distance mapping will not be considered in the lost function. In general, "
                           "adding the cosine distance mapping to the cost function will lead to more discriminative "
                           "embeddings, at the expense of having possible visual artefacts. By default it is set to "
                           "1.0 consider it in the cost function.")
        form.addParam('dimensions', IntParam, default=3,
                      label="Reduced landscape space dimensions?",
                      help="Determine the number of dimensions the reduced landscape will have. It must be a value "
                           "larger than three and smaller or equal to the number of dimension in the original space.")
        form.addParallelSection(threads=4, mpi=0)


    # --------------------------- INSERT steps functions ----------------------
    def _insertAllSteps(self):
        self._insertFunctionStep(self.computeReducedSpace)
        self._insertFunctionStep(self.createOutputStep)

    def createOutputStep(self):
        file_coords = self._getExtraPath("red_coords.txt")
        red_space = np.loadtxt(file_coords)

        inputSet = self.particles.get()
        progName = inputSet.getFlexInfo().getProgName()
        partSet = self._createSetOfParticlesFlex(progName=progName)

        partSet.copyInfo(inputSet)
        partSet.setHasCTF(inputSet.hasCTF())
        partSet.setAlignmentProj()

        idx = 0
        for particle in inputSet.iterItems():
            outParticle = ParticleFlex(progName=progName)
            outParticle.copyInfo(particle)

            outParticle.setZRed(red_space[idx])

            partSet.append(outParticle)

            idx += 1

        if self.mode.get() == 0:
            partSet.getFlexInfo().setAttr("umap_weights", self._getExtraPath("trained_umap"))

        self._defineOutputs(outputParticles=partSet)
        self._defineTransformRelation(self.particles, partSet)

    # --------------------------- STEPS functions -----------------------------
    def computeReducedSpace(self):
        particles = self.particles.get()
        self.num_vol = 0

        # ********* Get Z space *********
        z_space = []
        for particle in particles.iterItems():
            z_space.append(particle.getZFlex())
        z_space = np.asarray(z_space)
        # ********************

        # Generate files to call command line
        file_z_space = self._getExtraPath("z_space.txt")
        np.savetxt(file_z_space, z_space)

        # Compute reduced space
        file_coords = self._getExtraPath("red_coords.txt")
        mode = self.mode.get()
        if mode == 0:
            args = "--input %s --umap --output %s --n_epochs %d --n_components %d " \
                   % (file_z_space, file_coords, self.epochs_umap.get(), self.dimensions.get())

            if self.useGpu.get():
                gpu_list = ','.join([str(elem) for elem in self.getGpuList()])
                args += " --gpu %s" % gpu_list

            program = "dimensionality_reduction.py"
            program = flexutils.Plugin.getProgram(program)
            self.runJob(program, args)
        elif mode == 1:
            args = "--input %s --pca --n_components %d --output %s" \
                   % (file_z_space, self.dimensions.get(), file_coords)
            program = "dimensionality_reduction.py"
            program = flexutils.Plugin.getProgram(program)
            self.runJob(program, args)
        elif mode == 2:
            args = "--space %s --output %s --split_train 1 --clusters %d --init_power %f " \
                   "--end_power %f --vae_sigma %f --lat_dim %d --loss_lambda %f" \
                    % (file_z_space, file_coords, self.clusters.get(), self.init_power.get(),
                       self.end_power.get(), self.vae_sigma.get(),
                       self.dimensions.get(), self.loss_lambda.get())

            if self.useGpu.get():
                gpu_list = ','.join([str(elem) for elem in self.getGpuList()])
                args += " --gpu %s" % gpu_list

            program = flexutils.Plugin.getTensorflowProgram("train_deep_elastic.py", python=False)
            self.runJob(program, args, numberOfMpi=1)

    # --------------------------- INFO functions -----------------------------
    def _summary(self):
        summary = []
        if self.getOutputsSize() >= 1:
            for _, outParticles in self.iterOutputAttributes():
                summary.append("Output *%s*:" % outParticles.getNameId().split('.')[1])
        else:
            summary.append("Output particles not ready yet")
        return summary

    def _methods(self):
        return [
            "Dimensionality reduction of spaces based on different methods",
        ]

    def _validate(self):
        errors = []

        # Check number of reduced dimensions
        particles = self.particles.get()
        original_dimensions = len(particles.getFirstItem().getZFlex())
        dimensions = self.dimensions.get()
        if dimensions < 3:
            errors.append(f"The number of dimensions in the reduced space must be larger than 3. Currently "
                          f"it is set to {dimensions}")
        if dimensions > original_dimensions:
            errors.append(f"The number of dimensions in the reduced space must be smalller than or equal to "
                          f"the number of dimensions in the original space (original dimensions "
                          f"= {original_dimensions}. Currently set to {dimensions}."
                          "it is set to {dimensions}")

        return errors




# **************************************************************************
# *
# * Authors:     David Herreros Calero (dherreros@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************


import pyworkflow.viewer as pwviewer

import pwem.viewers.views as vi
from pwem.viewers import DataViewer, MODE, MODE_MD

import pwem.objects.data_flexhub as pwobj


class FlexDataViewer(pwviewer.Viewer):
    """
    Wrapper to visualize different type of objects
    with the Xmipp program xmipp_showj
    """
    _environments = [pwviewer.DESKTOP_TKINTER]
    _targets = [
        pwobj.SetOfClassesFlex,
        pwobj.SetOfClassesStructFlex
    ]

    def __init__(self, **kwargs):
        pwviewer.Viewer.__init__(self, **kwargs)
        self._views = []

    def _getObjView(self, obj, fn, viewParams={}):
        return vi.ObjectView(
            self._project, obj.strId(), fn, viewParams=viewParams)

    def _visualize(self, obj, **kwargs):
        views = []
        cls = type(obj)

        if issubclass(cls, pwobj.SetOfClassesFlex):
            views.append(vi.Classes3DView(self._project, obj.strId(),
                                          obj.getFileName()))

        elif issubclass(cls, pwobj.SetOfClassesStructFlex):
            views.append(vi.Classes3DView(self._project, obj.strId(),
                                          obj.getFileName()))

        return views


# Register specific sets in pwem dataviewer.
if hasattr(DataViewer, "registerConfig"):
    DataViewer.registerConfig(pwobj.SetOfParticlesFlex,
                              config={MODE: MODE_MD})
    DataViewer.registerConfig(pwobj.SetOfVolumesFlex,
                              config={MODE: MODE_MD})
    DataViewer.registerConfig(pwobj.SetOfAtomStructFlex,
                              config={MODE: MODE_MD})




# **************************************************************************
# *
# * Authors:     David Herreros Calero (dherreos@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************


import numpy as np
import os
from xmipp_metadata.metadata import XmippMetaData

import pyworkflow.protocol.params as params
from pyworkflow.object import Integer, String, Float, Boolean
from pyworkflow import VERSION_2_0

from pwem.protocols import ProtAnalysis3D, ProtFlexBase
import pwem.emlib.metadata as md
from pwem.constants import ALIGN_PROJ
from pwem.objects import ParticleFlex, SetOfParticlesFlex

from xmipp3.convert import writeSetOfImages, imageToRow, coordinateToRow, setXmippAttributes, createItemMatrix, \
    matrixFromGeometry

import flexutils.constants as const
import flexutils
from flexutils.utils import getXmippFileName


class XmippProtReassignReferenceZernike3D(ProtAnalysis3D, ProtFlexBase):
    """ Assignation of heterogeneity priors based on the Zernike3D basis. """
    _label = 'reassign reference - Zernike3D'
    _lastUpdateVersion = VERSION_2_0

    # --------------------------- DEFINE param functions --------------------------------------------
    def _defineParams(self, form):
        form.addSection(label='Input')
        form.addParam('inputParticles', params.PointerParam, label="Input particles", pointerClass='SetOfParticlesFlex')
        form.addParam('inputVolume', params.PointerParam, label="New reference", pointerClass='VolumeFlex',
                      help="Particles will be reassigned so this map is their new reference")
        form.addParam('mask', params.PointerParam, label="Reference mask", pointerClass='VolumeMask',
                      help="Mask associated to the new reference map")
        form.addParam('L1', params.IntParam, label="Zernike degree", expertLevel=params.LEVEL_ADVANCED,
                      default=3,
                      help="Zernike polynomial degree for the new focused Zernike3D coefficients")
        form.addParam('L2', params.IntParam, label="Spherical harmonic degree", expertLevel=params.LEVEL_ADVANCED,
                      default=2,
                      help="Spherical harmonics degree for the new focused Zernike3D coefficients")
        form.addParallelSection(threads=4, mpi=0)

    # --------------------------- INSERT steps functions --------------------------------------------
    def _insertAllSteps(self):
        self._insertFunctionStep(self.computeZernikeStep)
        self._insertFunctionStep(self.createOutputStep)

    # --------------------------- STEPS functions ---------------------------------------------------
    def computeZernikeStep(self):
        imgsFn = self._getExtraPath('inputParticles.xmd')
        particles = self.inputParticles.get()
        refMask = particles.getFlexInfo().refMask.get()
        prevL1 = particles.getFlexInfo().L1.get()
        prevL2 = particles.getFlexInfo().L2.get()
        Rmax = particles.getFlexInfo().Rmax.get()

        newRefMap = self.inputVolume.get()
        newRefMask = self.mask.get().getFileName()
        L1 = self.L1.get()
        L2 = self.L2.get()

        z_clnm_vec = {}
        # deformation_vec = {}
        for particle in particles.iterItems():
            z_clnm = particle.getZFlex()
            z_clnm_vec[particle.getObjId()] = z_clnm.reshape(-1)
            # deformation_vec[particle.getObjId()] = particle._xmipp_sphDeformation.get()

        def zernikeRow(part, partRow, **kwargs):
            imageToRow(part, partRow, md.MDL_IMAGE, **kwargs)
            coord = part.getCoordinate()
            idx = part.getObjId()
            if coord is not None:
                coordinateToRow(coord, partRow, copyId=False)
            if part.hasMicId():
                partRow.setValue(md.MDL_MICROGRAPH_ID, int(part.getMicId()))
                partRow.setValue(md.MDL_MICROGRAPH, str(part.getMicId()))
            partRow.setValue(md.MDL_SPH_COEFFICIENTS, z_clnm_vec[idx].tolist())
            # partRow.setValue(md.MDL_SPH_DEFORMATION, deformation_vec[idx])

        writeSetOfImages(particles, imgsFn, zernikeRow)

        # Write Zernike3D priors to file
        file_zclnm_r = self._getExtraPath('z_clnm_r,txt')
        with open(file_zclnm_r, 'w') as f:
            f.write(' '.join(map(str, [L1, L2, newRefMap.getFlexInfo().Rmax.get()])) + "\n")
            z_clnm = newRefMap.getZFlex()
            f.write(' '.join(map(str, z_clnm.reshape(-1))) + "\n")

        args = "--i %s --maski %s --maskr %s --zclnm_r %s --prevl1 %d --prevl2 %d --l1 %d --l2 %d --rmax %f --thr %d" \
               % (imgsFn, getXmippFileName(refMask), getXmippFileName(newRefMask),
                  file_zclnm_r, prevL1, prevL2, L1, L2, Rmax, self.numberOfThreads.get())
        program = "reassign_reference.py"
        program = flexutils.Plugin.getProgram(program)
        self.runJob(program, args)

    def createOutputStep(self):
        inputSet = self.inputParticles.get()
        partSet = self._createSetOfParticlesFlex(progName=const.ZERNIKE3D)
        mdOut = XmippMetaData(self._getExtraPath("inputParticles_reassigned.xmd"))

        partSet.copyInfo(inputSet)
        partSet.setHasCTF(inputSet.hasCTF())
        partSet.setAlignmentProj()

        coeffs = np.asarray([np.fromstring(item, sep=',') for item in mdOut[:, "sphCoefficients"]])
        deformation = mdOut[:, "sphDeformation"]
        shifts = mdOut[:, ["shiftX", "shiftY", "shiftZ"]]
        angles = mdOut[:, ["angleRot", "angleTilt", "anglePsi"]]

        inverseTransform = partSet.getAlignment() == ALIGN_PROJ

        idx = 0
        for particle in inputSet.iterItems():
            outParticle = ParticleFlex(progName=const.ZERNIKE3D)
            outParticle.copyInfo(particle)

            outParticle.setZFlex(coeffs[idx])
            outParticle.getFlexInfo().deformation = Float(deformation[idx])

            # Set new transformation matrix
            tr = matrixFromGeometry(shifts[idx], angles[idx], inverseTransform)
            outParticle.getTransform().setMatrix(tr)

            partSet.append(outParticle)

            idx += 1

        partSet.getFlexInfo().L1 = Integer(self.L1.get())
        partSet.getFlexInfo().L2 = Integer(self.L2.get())
        partSet.getFlexInfo().Rmax = inputSet.getFlexInfo().Rmax
        partSet.getFlexInfo().refMask = String(self.mask.get().getFileName())
        partSet.getFlexInfo().refMap = String(self.inputVolume.get().getFileName())

        self._defineOutputs(outputParticles=partSet)
        self._defineTransformRelation(self.inputParticles, partSet)

    # --------------------------- UTILS functions --------------------------------------------

    # ----------------------- VALIDATE functions ----------------------------------------
    def validate(self):
        """ Try to find errors on define params. """
        errors = []
        if isinstance(self.inputParticles.get(), SetOfParticlesFlex):
            if self.inputParticles.get().getFlexInfo().getProgName() != const.ZERNIKE3D:
                errors.append("The flexibility information associated with the particles is not "
                              "coming from the Zernike3D algorithm. Please, provide a set of particles "
                              "with the correct flexibility information.")
        return errors








# **************************************************************************
# *
# * Authors:     David Herreros Calero (dherreros@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************


import os

from pyworkflow.protocol.params import LabelParam, EnumParam
from pyworkflow.viewer import DESKTOP_TKINTER
from pwem.viewers import EmProtocolViewer

from flexutils.protocols.protocol_find_optimal_clusters import ProtFlexOptimalClusters


class FlexShowOptimalClustersViewer(EmProtocolViewer):
    """ Visualization of optimal cluster results. """

    _environments = [DESKTOP_TKINTER]
    _targets = [ProtFlexOptimalClusters]
    _label = 'analyze results'
    _method = ["Best clusters", "Calinski Harabasz", "Davies Bouldin",
               "Elbow", "Gap Statistic", "Silhouette"]

    def __init__(self, **kwargs):
        EmProtocolViewer.__init__(self, **kwargs)

    def _createFilenameTemplates(self):
        """ Centralize how files are called. """

        def _out(f):
            return self.protocol._getExtraPath(f)

        self._updateFilenamesDict({
            'Best clusters': _out('auto_clustering_results.png'),
            'Calinski Harabasz': _out('calinski_harabasz.png'),
            'Davies Bouldin': _out('davies_bouldin.png'),
            'Elbow': _out('elbow.png'),
            'Gap Statistic': _out('gap_statistic.png'),
            'Silhouette': _out('silhouette.png')
        })

    def _defineParams(self, form):
        form.addSection(label='Visualization')
        form.addParam('method', EnumParam,
                      choices=self._method,
                      default=0,
                      display=EnumParam.DISPLAY_COMBO,
                      label='Display results for analysis',
                      help="- Best clusters: Shows a scatter plot with the optimal clusters"
                           "found by each algorithm\n"
                           "- Calinski Harabasz: Shows the Calinski Harabasz plot (look for maxima)\n"
                           "- Davies Bouldin: Shows the Davies Bouldin plot (look for maxima)\n"
                           "- Elbow: Shows the Elbow plot (look for points close to the elbow)\n"
                           "- Gap Statistic: Shows the Gap Statistic plot (look for maxima)\n"
                           "- Silhouette: Shows the Silhouette plot (look for maxima)")
        form.addParam('doShowAnalysis', LabelParam,
                      label="Show analysis result")

    def _getVisualizeDict(self):
        self._createFilenameTemplates()

        visDict = {
            'doShowAnalysis': self._doShowAnalysis
        }

        return visDict

    def _doShowAnalysis(self, param=None):
        import matplotlib.image as mpimg
        import matplotlib.pyplot as plt
        fn = self._getFileName(self._method[self.method.get()])
        if os.path.exists(fn):
            img = mpimg.imread(fn)
            imgplot = plt.imshow(img)
            plt.axis('off')
            plt.show()
            return [imgplot]
        else:
            self.showError('File %s not found! Have you run analysis?' % fn)



# **************************************************************************
# *
# * Authors:     David Herreros Calero (dherreros@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************


import numpy as np
import pyvista as pv
import pyvistaqt as pvqt


class pointCloudPlot(object):

    def __init__(self, points=None, deformation=None):
        self.points = points[:, :3]
        self.deformation = deformation

        # Define actors
        self.points_actor = None

        # Define PyVista objects
        self.pv_points = pv.PolyData(self.points)
        self.pv_points["deformation"] = self.deformation
        self.pv_points["clusters"] = np.zeros(self.points.shape[0])

        self.plt = pvqt.BackgroundPlotter(title='Embedding 3D interactive viewer')
        self.plt.main_menu.clear()

        # Slider
        def changeOpacity(value):
            if self.points_actor is not None:
                self.points_actor.GetProperty().SetOpacity(value / 100)

        self.plt.add_slider_widget(changeOpacity, [0, 100], value=50,
                                   pointa=(.6, .9), pointb=(.9, .9),
                                   title='Embedding opacity', style='modern')

    # Display PC
    def plotPoints(self):
        self.points_actor = self.plt.add_mesh(self.pv_points, show_scalar_bar=False, scalars="deformation",
                                              cmap="gist_rainbow_r", opacity=0.5,
                                              render_points_as_spheres=True, reset_camera=True)

    def initializePlot(self):
        self.plotPoints()
        self.plt.app.exec_()


if __name__ == '__main__':
    import argparse

    # Input parameters
    parser = argparse.ArgumentParser()
    parser.add_argument('--coords', type=str, required=True)
    parser.add_argument('--deformation', type=str, required=True)

    args = parser.parse_args()

    # Read and generate data
    coords = np.loadtxt(args.coords)
    deformation = np.loadtxt(args.deformation)

    # Initialize volume slicer
    m = pointCloudPlot(points=coords, deformation=deformation)
    m.initializePlot()



# **************************************************************************
# *
# * Authors:     David Herreros Calero (dherreros@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************


import os
import re
import numpy as np
from glob import glob
from sklearn.neighbors import KDTree
from xmipp_metadata.image_handler import ImageHandler

from pyworkflow import NEW
from pyworkflow.protocol import LEVEL_ADVANCED
from pyworkflow.protocol.params import (PointerParam, IntParam, MultiPointerParam, BooleanParam, StringParam,
                                        USE_GPU, GPU_LIST)
import pyworkflow.utils as pwutils
from pyworkflow.utils.properties import Message
from pyworkflow.gui.dialog import askYesNo
from pyworkflow.object import Boolean, Integer

from pwem.protocols import ProtAnalysis3D, ProtFlexBase
from pwem.objects import SetOfVolumesFlex, VolumeFlex

import flexutils
from flexutils.utils import getOutputSuffix, computeNormRows
import flexutils.constants as const

import xmipp3


class ProtFlexAnnotateSpace(ProtAnalysis3D, ProtFlexBase):
    """ Interactive annotation of conformational spaces """

    _label = 'annotate space'
    _devStatus = NEW
    OUTPUT_PREFIX = 'flexible3DClasses'

    # --------------------------- DEFINE param functions ----------------------
    def _defineParams(self, form):
        form.addSection(label='General parameters')
        form.addHidden(USE_GPU, BooleanParam, default=True,
                       label="Use GPU for execution",
                       help="This protocol has both CPU and GPU implementation.\
                                     Select the one you want to use.")
        form.addHidden(GPU_LIST, StringParam, default='0',
                       expertLevel=LEVEL_ADVANCED,
                       label="Choose GPU IDs",
                       help="Add a list of GPU devices that can be used")
        form.addParam('particles', PointerParam, label="Particles to annotate",
                      pointerClass='SetOfParticlesFlex', important=True,
                      help="Particles must have a flexibility information associated (Zernike3D, CryoDrgn...")
        form.addParam('priors', MultiPointerParam, label="Priors", allowsNull=True,
                      pointerClass="SetOfVolumesFlex, VolumeFlex",
                      condition="particles and particles.getFlexInfo().getProgName() == 'Zernike3D'",
                      help='Volumes with Zernike3D coefficients associated (computed using '
                           '"Refernce map" as reference) to add as prior information to the Zernike3D '
                           'space')
        form.addParam('boxSize', IntParam, label="Box size",
                      condition="particles and particles.getFlexInfo().getProgName() == 'CryoDRGN'",
                      help="Volumes generated from the CryoDrgn network will be resampled to the "
                           "chosen box size (only for the visualization).")
        form.addParam('neighbors', IntParam, label="Number of particles to associate to selections",
                      default=5000, expertLevel=LEVEL_ADVANCED)



    # --------------------------- INSERT steps functions ----------------------
    def _insertAllSteps(self):
        self._insertFunctionStep(self.launchVolumeSlicer, interactive=True)

    def _createOutput(self):
        particles = self.particles.get()
        partIds = list(particles.getIdSet())
        neighbors = self.neighbors.get()
        num_part = particles.getSize()
        sr = particles.getSamplingRate()
        progName = particles.getFlexInfo().getProgName()

        # Get right imports
        if progName == const.NMA:
            createFn = self._createSetOfClassesStructFlex
            from pwem.objects import ClassStructFlex as Class
            from pwem.objects import AtomStructFlex as Rep
            from pwem.objects import SetOfClassesStructFlex as SetOfClasses
        else:
            createFn = self._createSetOfClassesFlex
            from pwem.objects import ClassFlex as Class
            from pwem.objects import VolumeFlex as Rep
            from pwem.objects import SetOfClassesFlex as SetOfClasses

        # Create SetOfFlexClasses
        suffix = getOutputSuffix(self, SetOfClasses)
        flexClasses = createFn(particles, suffix, progName=progName)

        # Read selected coefficients
        clInx = 1
        for file in glob(self._getExtraPath('saved_selections*')):
            z_space_vw = []
            with open(file) as f:
                lines = f.readlines()
                for line in lines:
                    z_space_vw.append(np.fromstring(line, dtype=float, sep=' '))
            z_space_vw = np.asarray(z_space_vw)
            # z_space_vw = np.asarray(z_space_vw[self.num_vol:])

            if "_cluster" in file:
                # Read space
                z_space = z_space_vw[1:]
                z_space_vw = z_space_vw[0][None, ...]
            else:
                # Read space
                z_space = np.loadtxt(self._getExtraPath("z_space.txt"))

                # Create KDTree
                kdtree = KDTree(z_space)

            # Popoulate SetOfClasses3D with KMean particles
            for z_idx in range(z_space_vw.shape[0]):
                if "_cluster" in file:
                    currIds = range(z_space.shape[0])
                    classId = [int(x) + 1 for x in re.findall('\d+', file)][-1]
                else:
                    _, currIds = kdtree.query(z_space_vw[z_idx].reshape(1, -1), k=neighbors + 10)
                    currIds = currIds[0]
                    classId = None

                newClass = Class()
                newClass.copyInfo(particles)
                newClass.setHasCTF(particles.hasCTF())
                newClass.setAcquisition(particles.getAcquisition())
                representative = Rep(progName=progName)
                if hasattr(representative, "setSamplingRate"):
                    representative.setSamplingRate(sr)

                # ****** Fill representative information *******
                if particles.getFlexInfo().getProgName() == const.ZERNIKE3D:
                    reference = particles.getFlexInfo().refMap.get()

                    # Resize coefficients
                    # factor = (ImageHandler().read(reference).getDimensions()[0] / 64)
                    # z_space_vw[z_idx] *= factor

                    representative.setLocation(reference)

                elif particles.getFlexInfo().getProgName() == const.CRYODRGN:
                    from cryodrgn.utils import generateVolumes
                    generateVolumes(z_space_vw[z_idx], particles.getFlexInfo()._cryodrgnWeights.get(),
                                    particles.getFlexInfo()._cryodrgnConfig.get(), self._getExtraPath(),
                                    downsample=self.boxSize.get(), apix=particles.getSamplingRate())
                    ImageHandler().scaleSplines(self._getExtraPath('vol_000.mrc'),
                                                self._getExtraPath('class_%d.mrc') % classId if classId else clInx,
                                                finalDimension=particles.getXDim(), overwrite=True)
                    representative.setLocation(self._getExtraPath('class_%d.mrc') % classId if classId else clInx)

                elif particles.getFlexInfo().getProgName() == const.HETSIREN:
                    from flexutils.utils import generateVolumesHetSIREN
                    generateVolumesHetSIREN(particles.getFlexInfo().modelPath.get(), z_space_vw[z_idx],
                                            self._getExtraPath(), step=particles.getFlexInfo().coordStep.get(),
                                            architecture=particles.getFlexInfo().architecture.get())
                    ImageHandler().scaleSplines(self._getExtraPath('decoded_map_class_01.mrc'),
                                                self._getExtraPath('class_%d.mrc') % classId if classId else clInx,
                                                finalDimension=particles.getXDim(), overwrite=True)
                    representative.setLocation(self._getExtraPath('class_%d.mrc') % classId if classId else clInx)

                elif particles.getFlexInfo().getProgName() == const.NMA:
                    reference = particles.getFlexInfo().refStruct.get()
                    subset = particles.getFlexInfo().atomSubset

                    representative.getFlexInfo().atomSubset = subset
                    representative.setLocation(reference)

                representative.setZFlex(z_space_vw[z_idx])
                representative.getFlexInfo().copyInfo(particles.getFlexInfo())
                # ********************

                newClass.setRepresentative(representative)

                flexClasses.append(newClass)

                enabledClass = flexClasses[newClass.getObjId()]
                enabledClass.enableAppend()

                if "_cluster" in file:
                    for itemId in currIds:
                        item = particles[partIds[itemId]]
                        item._xmipp_subtomo_labels = Integer(clInx)
                        enabledClass.append(item)
                else:
                    for idx in range(neighbors):
                        itemId = currIds[idx]
                        while itemId >= num_part:
                            currIds = np.delete(currIds, idx)
                            itemId = currIds[idx]
                        item = particles[partIds[itemId]]
                        item._xmipp_subtomo_labels = Integer(clInx)
                        enabledClass.append(item)

                flexClasses.update(enabledClass)
                clInx += 1

        # Save new output
        name = self.OUTPUT_PREFIX + suffix
        args = {}
        args[name] = flexClasses
        self._defineOutputs(**args)
        self._defineSourceRelation(particles, flexClasses)

    # --------------------------- STEPS functions -----------------------------
    def launchVolumeSlicer(self):
        particles = self.particles.get()
        self.num_vol = 0

        # ********* Get Z space *********
        z_space = []
        for particle in particles.iterItems():
            z_space.append(particle.getZFlex())
        z_space = np.asarray(z_space)

        if particles.getFlexInfo().getProgName() == const.ZERNIKE3D:
            reference = particles.getFlexInfo().refMap.get()
            mask = particles.getFlexInfo().refMask.get()
            volumes = self.priors

            # Copy original reference and mask to extra
            ih = ImageHandler()
            ih.convert(reference, self._getExtraPath("reference_original.mrc"), overwrite=True)
            ih.convert(mask, self._getExtraPath("mask_reference_original.mrc"), overwrite=True)

            # Resize reference map to increase real time conformation inspection performance
            inputFile = reference
            outFile = self._getExtraPath('reference.mrc')
            if not os.path.isfile(outFile):
                if pwutils.getExt(inputFile) == ".mrc":
                    inputFile += ":mrc"
                self.runJob("xmipp_image_resize",
                            "-i %s -o %s --dim %d " % (inputFile,
                                                       outFile,
                                                       64), numberOfMpi=1, env=xmipp3.Plugin.getEnviron())

            # Resize mask
            inputFile = mask
            outFile = self._getExtraPath('mask.mrc')
            if not os.path.isfile(outFile):
                if pwutils.getExt(inputFile) == ".mrc":
                    inputFile += ":mrc"
                self.runJob("xmipp_image_resize",
                            "-i %s -o %s --dim %d " % (inputFile,
                                                       outFile,
                                                       64), numberOfMpi=1, env=xmipp3.Plugin.getEnviron())
                self.runJob("xmipp_transform_threshold",
                            "-i %s -o %s --select below 0.01 "
                            "--substitute binarize " % (outFile, outFile), numberOfMpi=1, env=xmipp3.Plugin.getEnviron())

            z_space_vol = []
            if volumes:
                for pointer in volumes:
                    item = pointer.get()
                    if isinstance(item, VolumeFlex):
                        z_space_vol.append(item.getZFlex())
                    elif isinstance(item, SetOfVolumesFlex):
                        for volume in item.iterItems():
                            z_space_vol.append(volume.getZFlex())
            z_space_vol = np.asarray(z_space_vol)
            file_z_vol = self._getExtraPath("z_space_vol.txt")
            np.savetxt(file_z_vol, z_space_vol)

            # Resize coefficients
            # z_space = (64 / ImageHandler().read(reference).getDimensions()[0]) * z_space
        # ********************

        # Generate files to call command line
        file_z_space = self._getExtraPath("z_space.txt")
        file_interp_val = self._getExtraPath("interp_val.txt")
        np.savetxt(file_z_space, z_space)

        # Compute/Read UMAP or PCA
        file_coords = self._getExtraPath("red_coords.txt")
        red_space = []
        for particle in particles.iterItems():
            red_space.append(particle.getZRed())
        red_space = np.asarray(red_space)
        np.savetxt(file_coords, red_space)

        # ********* Get interpolation value for coloring the space *********
        if particles.getFlexInfo().getProgName() == const.ZERNIKE3D:
            interp_val = computeNormRows(z_space)
        else:
            interp_val = np.ones([1, z_space.shape[0]])
        # *********

        # Generate files to call command line
        np.savetxt(file_interp_val, interp_val)
        path = os.path.abspath(self._getExtraPath())

        # ********* Run viewer *********
        needsPackages = None

        if particles.getFlexInfo().getProgName() == const.ZERNIKE3D:
            L1 = particles.getFlexInfo().L1.get()
            L2 = particles.getFlexInfo().L2.get()
            args = "--data %s --z_space %s --interp_val %s --path %s " \
                   "--L1 %d --L2 %d --boxsize 64 --mode Zernike3D" \
                   % (file_coords, file_z_space, file_interp_val, path,
                      L1, L2)
            if volumes:
                args += "--z_space_vol %s" % file_z_vol

        elif particles.getFlexInfo().getProgName() == const.CRYODRGN:
            import cryodrgn
            args = "--data %s --z_space %s --interp_val %s --path %s " \
                   "--weights %s --config %s --boxsize %d --sr %f --mode CryoDrgn --env_name %s" \
                   % (file_coords, file_z_space, file_interp_val, path,
                      particles.getFlexInfo()._cryodrgnWeights.get(),
                      particles.getFlexInfo()._cryodrgnConfig.get(), self.boxSize.get(),
                      particles.getSamplingRate(), cryodrgn.Plugin.getActivationCmd())

        elif particles.getFlexInfo().getProgName() == const.HETSIREN:
            args = "--data %s --z_space %s --interp_val %s --path %s " \
                   "--weights %s --step %d --architecture %s --sr %f --mode HetSIREN --env_name flexutils-tensorflow" \
                   % (file_coords, file_z_space, file_interp_val, path,
                      particles.getFlexInfo().modelPath.get(),
                      particles.getFlexInfo().coordStep.get(),
                      particles.getFlexInfo().architecture.get(),
                      particles.getSamplingRate())

        elif particles.getFlexInfo().getProgName() == const.NMA:
            args = "--data %s --z_space %s --interp_val %s --path %s " \
                   "--weights %s --sr %f --boxsize %d --mode NMA --env_name flexutils-tensorflow" \
                   % (file_coords, file_z_space, file_interp_val, path,
                      particles.getFlexInfo().modelPath.get(),
                      particles.getSamplingRate(), particles.getXDim())

        if hasattr(particles.getFlexInfo(), "umap_weights"):
            args += " --reduce umap --umap_weights %s" % particles.getFlexInfo().getAttr("umap_weights")
        else:
            args += " --reduce pca"

        env = pwutils.Environ(os.environ)
        if self.usesGpu():
            env["CUDA_VISIBLE_DEVICES"] = ','.join([str(elem) for elem in self.getGpuList()])
        else:
            env["CUDA_VISIBLE_DEVICES"] = ''

        env["NAPARI_ASYNC"] = "1"

        program = "viewer_interactive_3d.py"
        program = flexutils.Plugin.getProgram(program, needsPackages=needsPackages, chimera=True)
        self.runJob(program, args, env=env)

        # *********

        if len(glob(self._getExtraPath("saved_selections*"))) > 0 and \
           askYesNo(Message.TITLE_SAVE_OUTPUT, Message.LABEL_SAVE_OUTPUT, None):
            self._createOutput()

    # --------------------------- INFO functions -----------------------------
    def _summary(self):
        summary = []
        if self.getOutputsSize() >= 1:
            for _, outClasses in self.iterOutputAttributes():
                summary.append("Output *%s*:" % outClasses.getNameId().split('.')[1])
                summary.append("    * Total annotated classes: *%s*" % outClasses.getSize())
        else:
            summary.append("Output annotated classes not ready yet")
        return summary

    def _methods(self):
        return [
            "Interactive annotation of conformational spaces",
        ]

    # ----------------------- VALIDATE functions ----------------------------------------
    def validate(self):
        """ Try to find errors on define params. """
        errors = []
        particles = self.particles.get()
        if particles.getFirstItem().getZRed().size == 0:
            errors.append("Particles do not have any dimensionality reduced version of the flexibility information "
                          "they store. Please, used the dimensionality reduction protocol available "
                          "in Flexutils Plugin to generate a valid set of particles")
        if particles.getSize() < self.neighbors.get():
            errors.append("Number of particles to be associated with each selected state is larger than the "
                          "total number of particles in the dataset. Please, provide a smaller value "
                          "(Advanced parameter)")

        # Check CryoDRGN boxsize parameter is set as it is mandatory
        if particles.getFlexInfo().getProgName() == 'CryoDRGN' and self.boxSize.get() is None:
            errors.append("Boxsize parameter needs to be set to an integer value smaller than or equal "
                          "to the boxsize used internally to train the CryoDRGN network")

        return errors
